/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <sys/processor.h>

#include <spix_sparc.h>
#include <shade.h>
#include <shade_anal.h>
#include "getopt.h"

/*
 * Define a standard trace record.
 */
struct shade_trace_s {SHADE_TRACE};


const char	shadeuser_analversion[] = "cachesim5 " "v1.0.1";



#define alloc(n,t)	((t *) domalloc ((unsigned int)((n) * sizeof (t))))
#define allocz(n,t)	((t *) docalloc ((unsigned int)((n) * sizeof (t))))
#define percent(a,b)	(100. * ((double)a) / ((double)b))

#define align(a,b)	(((a) >> (b)) << (b))
#define sizemask(x)	((1LL << (x)) - 1LL)

#define MAXSIZE		32	 /* log2 of max cache reference operand size */
#define NLEVELS		16	/* max # of cache levels */
#define THDS_CHUNK	16     /* chunk for additional thread allocation */

#define R		0	/* read */
#define W		1	/* write */
#define CW              3       /* clean write */

typedef spix_addr_t	addr_t;
typedef spix_addr_t	mask_t;
typedef unsigned long long	count_t;

#define C_LRU		1	/* least recently used set replacement */
#define C_RANDOM	2	/* random set replacement */
#define C_VIKING	3	/* viking set replacement */
#define C_PLRU		4	/* Pseudo-LRU set replacement */
#define C_LFSR		5	/* Simple LFSR pseudo-random replacement */
#define C_NVLFSR	6	/* LFSR pseudo-random replacement with Next Victim bits */
#define C_NVCTR		7	/* Counter pseudo-random replacement with Next Victim bits */

typedef struct CACHE {
	count_t       c_nref[1+MAXSIZE][2];
				/* # of accesses [size][read,write] */
	count_t       c_nmiss[1+MAXSIZE][2];
				/* # of misses [size][read,write] */
	count_t	c_ndwback;	/* # of dirty write backs from this cache */
	count_t	c_ncwback;	/* # of clean write backs from this cache */
	count_t       c_nrw;		/* # 1st write hit after read (write back) */
	count_t       c_nbkinving;	/* # subblocks victimized -> back invalidates */
	count_t       c_nbkinved;	/* # subblocks back invalidated */
	count_t       c_nwinv;	/* # subblocks write invalidated */
	addr_t	**c_tags;	/* address tags [#sets][#lines] */
	mask_t	**c_valid;	/* (subblock) valid bit(s) [#sets][#lines] */
	mask_t	**c_dirty;	/* (subblock) modified bit(s) [#sets][#lines] */
	char	        **c_used;	/* block used [#sets][#lines] */
	mask_t	*c_mru;		/* mru bits for viking [#lines] */
	spix_uint32_t **c_lru;	/* lru set replacement counters [1+#sets][#lines] */
	char	        *c_plru;	/* pseudo-lru set replacement codes [#lines] */
	int	        c_seed;		/* Seed value for random-type algorithms */
	char	        *c_nxtvctm;	/* Next victim (for NVLFSR and NVCTR) */
	int	        c_nlines;	/* # cache lines */
	short	        c_bsize;	/* log2 of cache block size (bytes) */
	short	        c_bbsize;	/* log2 of cache subblock size (bytes) */
	short	        c_ssize;	/* log2 of set size (bytes) */
	short	        c_nsets;	/* # of sets; 1 for direct mapped */
	char	        c_replace;	/* set replacement: C_RANDOM, etc. */
	char	        c_wthru;	/* write through (else write back) */
	char	        c_walloc;	/* write allocate */
	char	        c_winv;		/* write invalidate (wt+nowa only) */
	char          c_cwb;          /* clean write back */
	char	        c_idu;		/* 'I' for instruction cache, etc. */
	char	        c_level;	/* level # in cache hierarchy */
	char	        c_isbkinv;	/* on some other caches bkinv list */
	struct CACHE  **c_bkinv;      /* caches to back invalidate (inclusion) */
	struct CACHE  *c_next;	/* next cache level */
} Cache;


static char	ihash_memsize[SPIX_SPARC_NIOP];
static count_t	nina;

typedef struct CPU_CACHE {
	Cache	icaches[NLEVELS];	/* instruction caches */
	Cache	dcaches[NLEVELS];	/* data caches */
} CpuCache;

CpuCache**	 cpu_caches;

static int	nlevels;	  /* # of cache levels */
static int	cpus_used;	  /* # of simulated cpu's, default is 1 per thread */
static int	cpus_allocated;	  /* allocated CpuCache structs */
static int	nthreads;	  /* # of active application threads */

/* 
 * Mode "-single-cpu" : simulate all threads on a single CPU 
 * default is one CPU per thread
 */
static int	single_cpu_mode = FALSE;  
/* 
 * Mode "-pcs" : output per-cpu statistics 
 */
static int	per_cpu_statistics = FALSE ; 

/*
 * The process ID of the first Shade process.  Trace output from all other
 * processes will be labeled as coming from a child process.
 */
static pid_t	pidFirst;

enum{
    CACHESIM5_OPT_D,
    CACHESIM5_OPT_I,
    CACHESIM5_OPT_PCS,
    CACHESIM5_OPT_SINGLE_CPU,
    CACHESIM5_OPT_U
};

const shade_options_t shadeuser_opts[]={
       {"-d", "Specifies data cache in cachespec.", NULL, STR_PARAM},
       {"-i", "Specifies instruction cache in cachespec.", NULL, STR_PARAM},
       {"-pcs", "Provides per-cpu cache statistics for multithreaded programms", NULL, NO_PARAM},
       {"-single_cpu", "Forces all threads to be simulated as though they were executed on a single CPU.", NULL, NO_PARAM },
       {"-u", "Specifies instruction and data cache in cachespec.", NULL, STR_PARAM},
       {0, "", NULL, NO_PARAM}
};    


/* Forward declarations */
static void		accesscache(Cache *, addr_t, int, int);
static void		accesscacheset(Cache *, int, addr_t);
static void		backinvalidatedcache(Cache *, addr_t, int);
static void		backinvalidatingcache(Cache *, int, addr_t);
static void		backinvalidatingcachesubblock(Cache *, addr_t);
static void		cleancacheline(Cache *, int, addr_t);
static void		invalidatecache(Cache *);
static spix_bool_t	newcache(Cache *, char *, int);
static void		newcachesets(Cache *);
static int		replacecacheset(Cache *, addr_t);
static void		resetcache(Cache *);
static void		resetcaches(void);
static void		resetcachesets(Cache *);
static void		clear_cachestats(void);
static void		clear_cachestat(Cache *);
static void		cachestats(FILE *, Cache *, Cache *, count_t);
static void		cachestatsc(FILE *, Cache *);
static void		cachestatsl(FILE *, Cache *, Cache *, count_t);
static void		cachestatsla(FILE *, count_t, count_t, count_t);
static void		cacheusestats(FILE *, Cache *);
static void *		domalloc(unsigned int);
static void *		docalloc(unsigned int);
static addr_t		getsize(char *, char **);
static int		ilog2(addr_t);
static void		init_memsize(void);
static char *		ordinal(int);
static char *		putsize(addr_t);

static int		reallocate_caches(int);
static CpuCache*	new_cpu(CpuCache*);
static CpuCache*	get_cpu(int);
static void             bind_lwps();

int shadeuser_initialize(
	int		argc,
	char **		argv,
	char **		envp)
{
	shade_iset_t *	piset;
	shade_tset_t *	ptset;
	spix_bool_t	ok;
	Cache *		c;
	int		a;
	int		i;
	int		id = 0;
	Cache*	       dcaches;
	Cache*	       icaches;
    int                j;
    int                cachesim5_optnum;
    shade_option_val_t cachesim5_opt_value;

	/* allocate array for holdind caches */
	cpu_caches = (CpuCache**)allocz(THDS_CHUNK, CpuCache*);
	cpus_allocated = THDS_CHUNK;
	cpu_caches[0] = (CpuCache*)allocz(1,CpuCache);
	cpus_used = 1;
	
	dcaches = cpu_caches[0]->dcaches;
	icaches = cpu_caches[0]->icaches;

	ok = TRUE;
	for (i = j = 1; i < argc; i++) {

        cachesim5_optnum = shade_getoptv(argc, argv, &i, &j, shadeuser_opts, &cachesim5_opt_value);
        
        switch(cachesim5_optnum){
            case CACHESIM5_OPT_D:
            case CACHESIM5_OPT_I:
				if (id && id != shadeuser_opts[cachesim5_optnum].name[1]) {
                    shade_error("unmatched -i/-d");
					ok = FALSE;
					break;
				}
				if (nlevels >= NLEVELS) {
					shade_error("too many cache levels");
					ok = FALSE;
					break;
				}
				c = (shadeuser_opts[cachesim5_optnum].name[1] == 'i' ? icaches : dcaches) + nlevels;
				if (!newcache(c, argv[i], nlevels)) {
					ok = FALSE;
					break;
				}
				if (id == shadeuser_opts[cachesim5_optnum].name[1]) {
					nlevels++;
					id = 0;
				}
				else	id = 'i' + 'd' - shadeuser_opts[cachesim5_optnum].name[1];
				break;

            case CACHESIM5_OPT_PCS:
				per_cpu_statistics = TRUE;
                break;

            case CACHESIM5_OPT_SINGLE_CPU:
				single_cpu_mode = TRUE;
				/* bind all lwp's to single processor */
				bind_lwps();
                break;

            case CACHESIM5_OPT_U:
				if (id) {
					shade_error("unmatched -i/-d");
					ok = FALSE;
					break;
				}
				if (nlevels >= NLEVELS) {
					shade_error("too many cache levels");
					ok = FALSE;
					break;
				}
                if (!newcache(c = icaches + nlevels, argv[i], nlevels)) {
					ok = FALSE;
					break;
				}
				dcaches[nlevels++] = *c;
				break;

            default:
                shade_usage();
                exit(1);
        }
    }

                
                
/*        if (argv[i][0] != '-') {
			shade_error("Unknown option '%s'", argv[i]);
			ok = FALSE;
		}
		else {
			switch (a = argv[i][1])
			{
			case 'd':
			case 'i':
				if (id && id != a) {
                    shade_error("unmatched -i/-d");
					ok = FALSE;
					break;
				}
				if (nlevels >= NLEVELS) {
					shade_error("too many cache levels");
					ok = FALSE;
					break;
				}
				c = (a == 'i' ? icaches : dcaches) + nlevels;
				if (!newcache(c, argv[i], nlevels)) {
					ok = FALSE;
					break;
				}
				if (id == a) {
					nlevels++;
					id = 0;
				}
				else	id = 'i' + 'd' - a;
				break;

			case 'u':
				if (id) {
					shade_error("unmatched -i/-d");
					ok = FALSE;
					break;
				}
				if (nlevels >= NLEVELS) {
					shade_error("too many cache levels");
					ok = FALSE;
					break;
				}
				if (!newcache(c = icaches + nlevels, argv[i], nlevels)) {
					ok = FALSE;
					break;
				}
				dcaches[nlevels++] = *c;
				break;
			       
			case 's':
				if (!strcmp(argv[i]+1,"single-cpu")) {
					single_cpu_mode = TRUE;
*/					/* bind all lwp's to single processor */
/*					bind_lwps();
				} else {
					shade_usage();
					return 1;
				}
				break;

			case 'p':
				if (!strcmp(argv[i]+1,"pcs"))
					per_cpu_statistics = TRUE;
				else {
					shade_usage();
					return 1;
				}
				break;

			default:
				shade_error("Unknown option '%s'", argv[i]);
				ok = FALSE;
				break;
			}
		}
	}
*/
	if (id) {
		shade_error("unmatched -i/-d");
		ok = FALSE;
	}
	if (nlevels <= 0) {
		shade_error("Must specify a cache description");
		ok = FALSE;
	}
	if (single_cpu_mode)
		per_cpu_statistics = FALSE;
	if (!ok) {
		shade_usage();
		return(1);
	}
	init_memsize();

	for (i = 0; i < nlevels-1; i++) {
		cpu_caches[0]->icaches[i].c_next = cpu_caches[0]->icaches + i + 1;
		cpu_caches[0]->dcaches[i].c_next = cpu_caches[0]->dcaches + i + 1;
	}
	

	/*
	 * Record the initial process ID.
	 */
	pidFirst = getpid();

	/*
	 * We want to be notified of fork() operations, so we know if the
	 * application spawns a child process.
	 */
	shade_setopt(SHADE_OPT_FORKNOTIFY);

	shade_trsize(sizeof(shade_trace_t));

	/*
	 * Trace the opcode, PC, and annulled flag for all instructions.
	 */
	piset = shade_iset_newclass(SHADE_ICLASS_ANY, -1);
	if (single_cpu_mode)
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, -1);
	else
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, -1);
	shade_trctl(piset, SHADE_TRI_ISSUED, ptset);
	/*
	 * Also trace the effective address for loads and stores.
	 */
	piset = shade_iset_newclass(SHADE_ICLASS_LOAD, SHADE_ICLASS_USTORE, -1);
	if (single_cpu_mode)
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, SHADE_TRCTL_EA, -1);
	else
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, SHADE_TRCTL_EA, -1);

	shade_trctl(piset, SHADE_TRI_ISSUED, ptset);

	/*
	 * Trace the effective address and taken bit for conditional stores.
	 */
	piset = shade_iset_newclass(SHADE_ICLASS_LOAD, SHADE_ICLASS_CSTORE, -1);
	if (single_cpu_mode)
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, SHADE_TRCTL_EA, SHADE_TRCTL_TAKEN, -1);
	else
		ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID, SHADE_TRCTL_PC,
				       SHADE_TRCTL_ANNULLED, SHADE_TRCTL_EA, SHADE_TRCTL_TAKEN, -1);
	shade_trctl(piset, SHADE_TRI_ISSUED, ptset);

	return(0);
}


int shadeuser_analyze(void)
{
	shade_trace_t	atr[8*1024];
	shade_status_t	status;
	unsigned	ntr;
	unsigned	i;
	pid_t		pid;
	static pid_t		pid_parent;
	unsigned	tid;
	int		val;
	int		ih;
	CpuCache*	caches;

	status = shade_appstatus(&val);

	switch (status) {
	case SHADE_STATUS_LOADED:
		resetcaches();
		
		/*
		 * Remember the parent process ID so we can tell if it forks.
		 */
		pid_parent = getpid();
		break;
	case SHADE_STATUS_FORKED:
		/*
		 * If the application forks, zero the statistics in the child.
		 * The child shouldn't inherit statistics from the parent.
		 */
		pid = getpid();
		if (pid != pid_parent) {
			pid_parent = pid;
			clear_cachestats();
		}
		break;
	}


	ntr = shade_run(atr, sizeof(atr)/sizeof(atr[0]));


	for (i = 0;  i < ntr;  i++) {
		if (single_cpu_mode) {
			/* single-cpu mode, no per-thread simulation */
			caches = cpu_caches[0];
		} else {
			/*
			 * Get the ID of the thread producing this trace record.
			 */
			tid = atr[i].tr_tid;
			caches = get_cpu(tid);
		}


		accesscache(caches->icaches, atr[i].tr_pc, R, 2);

		
				/*
				 * This is somewhat untrue. We usually
				 * speculate well past branches in modern
				 * pipelines. So this instruction most
				 * likely has been sent to the data caches.
				 * Sreeram 7/98
				 *
				 * Here purely for backward compatibility
				 * with older version of cachesim5.
				 */
		if (atr[i].tr_annulled)
			continue;
		else
			nina++;
		
				/* BUG: should check alignment */
		ih = atr[i].tr_ih;
		if (spix_sparc_iop_isload(ih)) {


			accesscache(caches->dcaches, atr[i].tr_ea, R,
				    ihash_memsize[ih]);



		}
		if (spix_sparc_iop_isustore(ih)) {

			accesscache(caches->dcaches, atr[i].tr_ea, W,
				    ihash_memsize[ih]);

		}
		else if (spix_sparc_iop_iscstore(ih) && atr[i].tr_taken)
		{

			accesscache(caches->dcaches, atr[i].tr_ea, W,
				    ihash_memsize[ih]);


		}
	}
	

	return(0);
}


void ic_merge(Cache* ic, Cache* ic_dest)
{
	int rw;
	int sz;
	int i;
	for (sz = 0; sz <= MAXSIZE; sz++)
		for (rw = 0; rw < 2; rw++) {
			ic_dest->c_nmiss[sz][rw] += ic->c_nmiss[sz][rw];
			ic_dest->c_nref[sz][rw] += ic->c_nref[sz][rw];
		}
	ic_dest->c_ndwback += ic->c_ndwback;
	ic_dest->c_ncwback += ic->c_ncwback;
	ic_dest->c_nrw += ic->c_nrw;
	ic_dest->c_nbkinving += ic->c_nbkinving;
	ic_dest->c_nbkinved += ic->c_nbkinved;
	ic_dest->c_nwinv += ic->c_nwinv;
}


void
shadeuser_report(int reason, void* ptr) {
	pid_t     	pid;
	int	i;
	Cache*	dc;
	Cache*	ic;
  
	switch (reason) {
	case SHADEUSER_REPORT_TERMINATE:
		/* no info is consolidated for applications */
		break;
	  
	case SHADEUSER_REPORT_SIGNAL:		
	case SHADEUSER_REPORT_APP_DONE:		
		/*
		 * If this is a child process, say so in the output.
		 */
		pid = getpid();
		if (pid != pidFirst)
			fprintf(shade_fp, "FROM CHILD PROCESS %d:\n\n", pid);
		
		fprintf(shade_fp, "Analyzer : %s\nCache configuration", shade_analversion);	    
		if (cpus_used > 1) {
			fprintf(shade_fp, " (%d CPU(s) used)", cpus_used);
			fprintf(shade_fp, per_cpu_statistics ? 
				", per cpu statistics:" : ":");	    
		}	  
		else
			fprintf(shade_fp, ":");	  
		
		
		if (per_cpu_statistics) {
			for (i=0; i<cpus_used; i++) {
				if (cpus_used > 1)
					fprintf(shade_fp, "CPU ID %d:\n", i);	    
				cachestats(shade_fp, cpu_caches[i]->icaches, cpu_caches[i]->dcaches, nina);	 
				for (ic=cpu_caches[i]->icaches, dc=cpu_caches[i]->dcaches; ic&&dc; 
				     ic=ic->c_next, dc=dc->c_next) {
					cacheusestats (shade_fp, ic);
					if (dc->c_idu == 'D')
						cacheusestats (shade_fp, dc);
				}
			}
		} else {
			/* Summarize statistics*/
			int i;
			int nlevel;
			/* Buffer cache data to summarize values 
			 * Static as we do nor really have free_cache procedure
			 */

			static CpuCache* buf = 0;
			if (buf) {
				for (nlevel = 0; nlevel < nlevels; nlevel++) {
					resetcache(&buf->icaches[nlevel]);
					resetcache(&buf->dcaches[nlevel]);
					/* clear statistics */
					clear_cachestat(&buf->icaches[nlevel]);	
					clear_cachestat(&buf->dcaches[nlevel]);	
				}
			} else
				buf = new_cpu(cpu_caches[0]);
						      
			/* merge all cpu data into buf */
			for (i=0; i<cpus_used; i++)
				for (nlevel = 0; nlevel < nlevels; nlevel++) {
					ic_merge(cpu_caches[i]->icaches+nlevel, 
						 buf->icaches+nlevel);
					ic_merge(cpu_caches[i]->dcaches+nlevel, 
						 buf->dcaches+nlevel);	      
				}
			
			/* print summary statistics */
			cachestats(shade_fp, buf->icaches, buf->dcaches, nina);
			
			for (i=0; i<cpus_used; i++) {
				/* we cannot summarize used blocks statistics */
				if (cpus_used > 1 ) 
					fprintf(shade_fp, "CPU %d CACHE USAGE STATISTICS:\n", i);
				for (ic=cpu_caches[i]->icaches, dc=cpu_caches[i]->dcaches; ic&&dc; 
				     ic=ic->c_next, dc=dc->c_next) {
					cacheusestats (shade_fp, ic);
					if (dc->c_idu == 'D')
						cacheusestats (shade_fp, dc);
				}
			}
		}
		/* flush the output stream */
		fflush(shade_fp);
		break;
		
	}	
	
}


int shadeuser_terminate(
	int	ret)
{
	int	val;

	/*
	 * Exit with the termination status from the executed application.
	 */
	switch (shade_appstatus(&val))
	{
	case SHADE_STATUS_EXITED:
		return(val);

	case SHADE_STATUS_SIGNALED:
		return(128 + val);

	case SHADE_STATUS_NOAPP:
		return(0);

	case SHADE_STATUS_LOADED:
	case SHADE_STATUS_RUNNING:
	default:
		assert(0);
	}

}


void shadeuser_analusage(void)
{
    /*
    fprintf(stderr, "-{idu}<sz>[b<bsz>[,<subbsz>]][s<assoc>][r<replace>]"
	    "[wb][wt][wa][wi][I<cache>] - cache parameters:\n");
    fprintf(stderr, "\t\t-{i|d|u}	select instruction, data, unified cache;\n");
    fprintf(stderr, "\t\t<sz>	overall cache size in bytes (suffix K for 1024x, etc.);\n");
    fprintf(stderr, "\t\tb<bsz>[,<bbsz>]  block and subblock sizes in bytes;\n");
    fprintf(stderr, "\t\t[s<#sets>] set associativity (1 for direct mapped (default));\n");
    fprintf(stderr, "\t\t[r<replace>]	set replacement: lru, random, viking, plru, lfsr, nvlfsr, nvctr;\n");

    fprintf(stderr, "\t\t[wt|wb]	write thru or write back;\n");
    fprintf(stderr, "\t\t[wa]		write allocate;\n");
    fprintf(stderr, "\t\t[wi]		write invalidate;\n");
    fprintf(stderr, "\t\t[I<cache>] include cache (e.g. i1 for 1st level I cache);\n");

    fprintf(stderr, "-single-cpu - \tsimulate all threads on a single cpu;\n");
    fprintf(stderr, "-pcs -	       \toutput per-cpu cache statistics;\n");*/

    shade_print_opt_info(shadeuser_opts, 0);
}


/*
 * access cache c for reading (rw=R), writing (rw=W), or
 * clean write back (rw=CW) 1<<size bytes at address a.
 * a is presumed to be aligned on 1<<size boundary
 * (size lsb's of a == 0).
 * For R -> rw = 0 (=R); for W/CW -> rw%2 = 1 (=W).
 */

static void accesscache(
	Cache * c,
	addr_t	a,
	int	rw,
	int	size)
{
	int	bsize, bbsize, ssize, s, rwalloc;
	addr_t	T, X;
	mask_t	Y;

top:

	ssize = c->c_ssize;	/* set size (2^ssize) */
	bsize = c->c_bsize;	/* block size (2^bsize)*/
	bbsize = c->c_bbsize;	/* subblock size (2^bbsize) */
                /*  (= block size if not subblocked) */

	rwalloc = rw == R || ((rw%2) == W && c->c_walloc);

	/* check if size of data unit bigger than subblock size */
	/*  if yes, break it up */
	if (size > bbsize) {
		addr_t	ainc, 
			alim = a + (1 << size); /* upper limit of addr */

		/* check if null cache (hole in cache system) */
		if (bbsize < 0) {
			c->c_nref[size][rw%2]++;
			/* if there is a higher cache level, forward access */

		if (c->c_next) {
				accesscache (c->c_next, a, rw, size);
                        }
			return; /* after returning, done */
		}


		/* access this same cache level as many times as subblock size */
		/*  fits in data unit size */
		for (ainc = 1 << bbsize; a < alim; a += ainc) {
			accesscache (c, a, rw, bbsize);
                }
		return; /* after returning, done */
	}

	c->c_nref[size][rw%2]++;

	/* +----------+----------+----------+----------+ */
	/* |   tag    |   index  | subblk # |  offset  | */
	/* +----------+----------+----------+----------+ */
	/*            |<-------------ssize------------>| */
	/*                       |<-------bsize------->| */
	/*                                  |<-bbsize->| */

	T = a >> ssize;				/* tag */
	X = (a & sizemask (ssize)) >> bsize;	/* block # (index) */
	Y = (a & sizemask (bsize)) >> bbsize;	/* subblock # */
	Y = 1 << Y;				/* subblock bit mask */

	/* search through all sets for a match */
	for (s = 0; s < c->c_nsets; s++)
		/* check if tag matches */
		if (c->c_tags[s][X] == T) {
			/* check if line is valid */
			if (c->c_valid[s][X] & Y) {
	        /* if set assoc., do replacement bookkeeping */
			  if (c->c_nsets > 1) {
					accesscacheset (c, s, X);
                          }
		/* for R or CW access, we are done */
			  if (rw != W) {
					return;
                          }
		/* for write access, not done yet... */
				else	goto hit;
			}
			/* if line is not valid */
			else {
		/*
		 * nonsubblocked caches don't distinguish tag
		 * mismatch misses from unset valid bit misses.
		 */
	        /* if not subblocked, check next set */
				if (bsize == bbsize)
					continue;
		/* otherwise, handle the miss */
				goto miss;
			}
		}

	/* if R & no vc or W & wa --> allocate */
	if (rwalloc) {
		/* find set to be replaced */
	  if (c->c_nsets > 1) {
			s = replacecacheset (c, X); /* for set-assoc. */
          }
		else	s = 0;                      /* for direct mapped */

		/* if any valid bits are set, the line may have to be evicted */
		if (c->c_valid[s][X]) {
			/* if it's not a wt cache or it's an R access to a clean */
			/*  wb cache, make room for the new line */
		  if ((!c->c_wthru) || ((c->c_cwb) && (rw == R))) {
				cleancacheline (c, s, X);
                  }

			/* for inclusive caches, back invalidate evicted line */
			if (c->c_bkinv)
				backinvalidatingcache (c, s, X);

			c->c_valid[s][X] = 0; /* clear the valid bits */
		}
		c->c_tags[s][X] = T; /* set the new tag */
	}
	else	/*
		 * on write miss to virtual, write thru, no write alloc cache,
		 * invalidate subblock when tag mismatches in case of aliasing.
		 * only implemented for 1st level, direct mapped caches.
		 */
		if (c->c_winv)
			if (c->c_valid[0][X] & Y) {
				c->c_valid[0][X] &= ~Y;    /* c_bkinv==0 */
				c->c_nwinv++;
			}

miss:
	c->c_nmiss[size][rw%2]++;

	/* check if line fill is required */
	if (c->c_next && ((rw%2) == W) && c->c_walloc)
		/* only do a line fill if size is smaller than */
		/*  subblock size of next cache */
	  if (size < bbsize) {
			accesscache (c->c_next, align (a, bbsize), R, bbsize);
          }

	if (rwalloc) {
		c->c_valid[s][X] |= Y;
		c->c_used[s][X] = 1;
		if (c->c_nsets > 1)
			accesscacheset (c, s, X);
	}

	/* if it's a W/CW and not a wt cache or */
	/*  it's a CW and a wt, wa cache, */
	/*  access does not go further */
	if ((((rw%2) == W) && !c->c_wthru) || ((rw == CW) && c->c_wthru && c->c_walloc)) {
		/* if it's a W, set the dirty bit */
		if (rw == W)
			c->c_dirty[s][X] |= Y;
	}
	/* if it's a R or a wt cache, propagate access further if possible */
	else	if (c->c_next)
		goto nextlevel;

	return;

hit:
	if (c->c_wthru) {
		if (c->c_next)
			goto nextlevel;
	}
	else {
		if (c->c_dirty[s][X] == 0)
			c->c_nrw++;
		c->c_dirty[s][X] |= Y;
	}

	return;

nextlevel:
	/*
	 * accesscache (c->c_next, align (a, bbsize), rw, bbsize);
	 * (tail recursion elimination)
	 */
	/* if this is a wt cache and a W access, keep the data access size */
	if (!c->c_wthru || (rw == R)) {
		a = align (a, bbsize);	/* in case bbsize > c->c_next->c_bbsize */
		size = bbsize;
	}
	c = c->c_next;
	goto top;
}


static void accesscacheset(
	Cache * c,
	int	s,
	addr_t	X)
{
	switch (c->c_replace) {
	case C_LRU:
		if (c->c_lru[s][X] = ++c->c_lru[-1][X])
			break;
				/*
				 * minor glitch on lru counter overflow ...
				 */
		for (s = -1; s < c->c_nsets; s++)
			c->c_lru[s][X] = 0;
		break;
	case C_RANDOM:
		break;
	case C_VIKING:
		c->c_mru[X] |= 1 << s;
		if (c->c_mru[X] == sizemask (c->c_nsets))
			c->c_mru[X] = 1 << s;
		break;
	case C_PLRU:
				/* Here's the trick: we use 3 bits.  The MSB indicates which
				 * half (0-1 vs. 2-3) was accessed more recently.  The middle
				 * bit indicates whether set 0 was accessed more recently than
				 * set 1.  The LSB indicates whether set 2 was accessed more
				 * recently than set 3. */
		switch (s) {
		case 0:
			c->c_plru[X] = c->c_plru[X] | 6;
			break;
		case 1:
			c->c_plru[X] = (c->c_plru[X] | 4) & 5;
			break;
		case 2:
			c->c_plru[X] = (c->c_plru[X] | 1) & 3;
			break;
		case 3:
			c->c_plru[X] = c->c_plru[X] & 2;
			break;
		}
		break;
	case C_LFSR:
		break;
	case C_NVLFSR:
		break;
	case C_NVCTR:
		break;
	}
}


/*
 * invalidate any data in cache c in address range [a,a+1<<size)
 */

static void backinvalidatedcache(
	Cache * c,
	addr_t	a,
	int	size)
{
	int	bsize, bbsize, ssize, s;
	addr_t	a2, ainc, ainc2, alim, T, X;
	mask_t	Y, v;

	ssize = c->c_ssize;	/* set size */
	bsize = c->c_bsize;	/* block size */
	bbsize = c->c_bbsize;	/* subblock size */

	T = a >> ssize;				/* tag */
	X = (a & sizemask (ssize)) >> bsize;	/* block # (index) */

	if (size < bsize)
		goto small;

	/*
	 * invalidate a block at a time.
	 */

	alim = a + (1 << size);
	ainc = 1 << bsize;
	ainc2 = 1 << bbsize;

	for (; a < alim; a += ainc, X++)
		for (s = 0; s < c->c_nsets; s++) {
			if (T != c->c_tags[s][X])
				continue;

			v = c->c_valid[s][X];

			for (a2 = a; a2 < a+ainc; a2 += ainc2, v >>= 1)
				if (v & 1) {
					c->c_nbkinved++;
					if (c->c_bkinv)
						backinvalidatingcachesubblock
							(c, a2);
				}

			c->c_valid[s][X] = 0;
		}
	return;
	
 small:
	/*
	 * invalidate a subblock at a time.
	 * inefficient below, but unlikely case.
	 */

	if (size < bbsize) {
		size = bbsize;			/* invalidate whole subblock */
		a = (a >> size) << size;	/* realign */
	}

	Y = (a & sizemask (bsize)) >> bbsize;	/* subblock # */
	Y = 1 << Y;				/* subblock bit mask */

	alim = a + (1 << size);
	ainc = 1 << bbsize;

	for (; a < alim; a += ainc, Y <<= 1)
		for (s = 0; s < c->c_nsets; s++)
			if (T == c->c_tags[s][X])
				if (c->c_valid[s][X] & Y) {
					c->c_nbkinved++;
					c->c_valid[s][X] &= ~Y;
					if (c->c_bkinv)
						backinvalidatingcachesubblock
							(c, a);
				}
	return;
}


/*
 * for each valid subblock in set s, line X of cache c
 * generate a back invalidate request to each cache
 * included by cache c.
 */

static void backinvalidatingcache(
	Cache * c,
	int	s,
	addr_t	X)
{
	addr_t	a, ainc, alim;
	mask_t	v;

	a = (c->c_tags[s][X] << c->c_ssize) | (X << c->c_bsize);
	ainc = 1 << c->c_bbsize;
	alim = a + (1 << c->c_bsize);

	v = c->c_valid[s][X];

	for (; a < alim; a += ainc, v >>= 1)
		if (v & 1)
			backinvalidatingcachesubblock (c, a);

}


static void backinvalidatingcachesubblock(
	Cache * c,
	addr_t	a)
{
	Cache	**bc;

	c->c_nbkinving++;

	for (bc = c->c_bkinv; *bc; bc++)
		backinvalidatedcache (*bc, a, c->c_bbsize);
}

/*
 * write back dirty subblocks in set s, line X of cache c.
 */

static void cleancacheline(
	Cache * c,
	int	s,
	addr_t	X)
{
	int	bsize, bbsize;
	addr_t	a, ainc, alim;
	mask_t	vdm, vcm, Y;

	vdm = c->c_valid[s][X] & c->c_dirty[s][X]; /* generate valid/dirty mask */

	vcm = 0;
	/* if clean wb cache, also writeback valid clean lines */
	if (c->c_cwb)
		vcm = c->c_valid[s][X] & ~(c->c_dirty[s][X]); /* generate valid/clean mask */

	c->c_dirty[s][X] = 0; /* clear dirty bits */

	/* if no dirty or clean wb required, return */
	if ((vdm == 0) && (vcm == 0))
		return;

        /*
         * This fragment has been missed somehow in V6 !!!
         */

	bsize = c->c_bsize;
	bbsize = c->c_bbsize;

	a = (c->c_tags[s][X] << c->c_ssize) | (X << bsize);
	alim = a + (1 << bsize);
	ainc = 1 << bbsize;


	/* go through all subblocks */
	for (Y = 1; a < alim; a += ainc, Y <<= 1) {
				/* check if dirty or clean wb is required */
		if ((vdm & Y) || (vcm & Y))
			/* if a dirty wb is required, ... */
			if (vdm & Y) {
				c->c_ndwback++; /* ...increment dirty counter, and... */
				/* ...if there is a higher cache level, do wb */
				if (c->c_next)
					accesscache (c->c_next, a, W, bbsize);
			}
			else { /* a clean wb is required, ... */
				c->c_ncwback++; /* ...increment clean counter, and... */
				/* ...if there is a higher cache level, do cwb */
				if (c->c_next)
					accesscache (c->c_next, a, CW, bbsize);
			}
	}
}


static void invalidatecache(
	Cache * c)
{
	int	s;
	addr_t	X;

	if (c->c_valid)
		for (s = 0; s < c->c_nsets; s++)
			for (X = 0; X < c->c_nlines; X++)
				c->c_valid[s][X] = 0;
 
	if (c->c_dirty)
		for (s = 0; s < c->c_nsets; s++)
			for (X = 0; X < c->c_nlines; X++)
				c->c_dirty[s][X] = 0;
}


/*
 * -{i|d|u}	select instruction, data, unified cache
 * <sz>		overall cache size in bytes (suffix K for 1024x, etc.)
 * b<bsz>[,<bbsz>]  block and subblock sizes in bytes
 * [s<#sets>]	set associativity (1 for direct mapped (default))
 * [r<replace>] set replacement: lru, random, plru, nvctr
 * [wt|wb]	write thru or write back
 * [wa]		write allocate
 * [wi]		write invalidate
 * [cw]         clean write back
 * [I<cache>]	include cache (e.g. i1 for 1st level I cache)
 */

static spix_bool_t newcache(
	Cache * c,
	char *	s,
	int  level )
{
	Cache	*bc[2*NLEVELS];
	char	*save = s;
	int	i, idu, nbc = 0;
	int	bsz = -1, bbsz = -1, ssz;
	int	nsets = 1, repl = -1;
	int	wbt = 0, wa = 0, wi = 0;
	addr_t	sz, t;
	int     cwb = 0;
	Cache*	dcaches;
	Cache*	icaches;
	
	dcaches = cpu_caches[0]->dcaches;
	icaches = cpu_caches[0]->icaches;

	c->c_level = level+1; /* starting from 1 */ 
	c->c_idu = s[1] + 'A' - 'a';
	sz = getsize (s+2, &s);
	if (sz <= 0) {
		if (*s) {
			shade_error("%s: bad size (%u)", save, sz);
			return(FALSE);
		}

				/*
				 * null cache
				 */
		c->c_nlines = 0;
		c->c_bsize = -1;
		c->c_bbsize = -1;
		c->c_ssize = -1;
		c->c_nsets = 0;
		c->c_replace = C_RANDOM;
		c->c_wthru = 1;
		c->c_walloc = 0;
		c->c_winv = 0;
		c->c_cwb = 0;
		c->c_isbkinv = 0;
		c->c_bkinv = 0;
		c->c_tags = 0;
		c->c_valid = 0;
		c->c_dirty = 0;
		return(TRUE);
	}

	while (*s)
		switch (*s) {
		case 'b':
			if (0 > (bsz = ilog2 (t = getsize (s+1, &s)))) {
				shade_error("%s: block size (%u) not power of 2",
					    save, t);
				return(FALSE);
			}
			if (*s != ',') {
				bbsz = bsz;
				break;
			}
			if (0 > (bbsz = ilog2 (t = getsize (s+1, &s)))) {
				shade_error("%s: subblock size (%u) not power of 2",
					    save, t);
				return(FALSE);
			}
			if (bsz < bbsz) {
				shade_error("%s: block size (%u) < subblock size (%u)", 
					    save, 1 << bsz, 1 << bbsz);
				return(FALSE);
			}
			break;
		case 'r':
			if (!strncmp (s+1, "lru", 3)) {
				repl = C_LRU;
				s += 4;
			}
			else if (!strncmp (s+1, "random", 6)) {
				repl = C_RANDOM;
				s += 7;
			}
			else if (!strncmp (s+1, "viking", 6)) {
				repl = C_VIKING;
				s += 7;
			}
			else if (!strncmp (s+1, "plru", 4)) {
				repl = C_PLRU;
				s += 5;
			}
			else if (!strncmp (s+1, "lfsr", 4)) {
				repl = C_LFSR;
				s += 5;
			}
			else if (!strncmp (s+1, "nvlfsr", 6)) {
				repl = C_NVLFSR;
				s += 7;
			}
			else if (!strncmp (s+1, "nvctr", 5)) {
				repl = C_NVCTR;
				s += 6;
			}
			else {
				shade_error("%s: bad set replacement", save);
				return(FALSE);
			}
			break;
		case 's':
			nsets = strtol (s+1, &s, 10);
			if (nsets <= 0) {
				shade_error("%s: bad set associativity", save);
				return(FALSE);
			}
			break;
		case 'w':
			switch (*++s) {
			case 'a':
				wa = *s;
				break;
			case 'b':
			case 't':
				wbt = *s;
				break;
			case 'i':
				wi = 1;
				break;
			default:
				shade_error("%s: w%c invalid", save, *s);
				return(FALSE);
			}
			s++;
			break;
		case 'c':
			switch (*++s) {
			case 'w':
				cwb = 1;
				break;
			default:
				shade_error ("%s: c%c invalid", save, *s);
				return(FALSE);
			}
			s++;
			break;
		case 'I':
			idu = s[1];
			i = strtol (s+2, &s, 10);

			if (i <= 0 || i > nlevels ||
			    idu != 'i' && idu != 'd' && idu != 'u' ||
			    (idu == 'u') ^ (icaches[i-1].c_idu == 'U') ||
			    nbc >= 2 * NLEVELS)
			{
				shade_error("%s: bad cache inclusion", save);
				return(FALSE);
			}

			if (idu == 'd')
				bc[nbc++] = dcaches + i - 1;
			else	bc[nbc++] = icaches + i - 1;	/* i or u */
			break;
		default:
			shade_error("%s: bad cache option (%s)", save, s);
			return(FALSE);
		}

	if (bsz < 0) {
		shade_error("%s: bad/missing block size", save);
		return(FALSE);
	}

	/*
	 * insure that a 32 bit word (mask_t) suffices
	 * to hold all of the subblock valid bits.
	 */
	if (bsz - bbsz > 5) {
		shade_error("%s: subblock size (%d) << block size (%d)",
			    save, 1 << bbsz, 1 << bsz);
		return(FALSE);
	}

	if (sz % nsets) {
		shade_error("%s: cache size (%s) not multiple of assoc (%d)",
			    save, putsize (sz), nsets);
		return(FALSE);
	}
	if (0 > (ssz = ilog2 (t = sz / nsets))) {
		shade_error("%s: set size (%s) not power of 2",
			    save, putsize (t));
		return(FALSE);
	}
	if (ssz < bsz) {
		shade_error("%s: set size (%s) < block size (%d)", 
			    save, putsize (t), 1 << bsz);
		return(FALSE);
	}

	if (!wbt)
		wbt = wa ? 'b' : 't';
	if (wbt == 'b')
		wa = 'a';	/* write back implies write allocate */

	c->c_bsize = bsz;
	c->c_bbsize = bbsz;
	c->c_nsets = nsets;
	c->c_ssize = ssz;
	c->c_nlines = 1 << (c->c_ssize - c->c_bsize);
	c->c_walloc = wa == 'a' ? 1 : 0;
	c->c_wthru = wbt == 't' ? 1 : 0;

	c->c_tags = alloc (c->c_nsets, addr_t *);
	c->c_valid = alloc (c->c_nsets, mask_t *);
	c->c_dirty = alloc (c->c_nsets, mask_t *);
	c->c_used = alloc (c->c_nsets, char *);

	for (i = 0; i < c->c_nsets; i++) {
		c->c_tags[i] = allocz (c->c_nlines, addr_t);
		c->c_valid[i] = allocz (c->c_nlines, mask_t);
		c->c_dirty[i] = allocz (c->c_nlines, mask_t);
		c->c_used[i] = allocz (c->c_nlines, char);
	}

	if (c->c_nsets > 1) {
		c->c_replace = repl < 0 ? C_RANDOM : repl;
				/*
				 * mru bit mask (1 bit per set) must fit in mask_t.
				 */
		if (c->c_replace == C_VIKING && c->c_nsets > 32) {
			shade_error("%s: set associativity too high for viking",
				    save);
			return(FALSE);
		}
		newcachesets (c);
	}
	else if (repl >= 0) {
		shade_error("%s: set replacement not allowed for direct mapped cache",
			    save);
		return(FALSE);
	}
		
	c->c_isbkinv = 0;

	if (nbc <= 0)
		c->c_bkinv = 0;
	else {
		c->c_bkinv = alloc (1 + nbc, Cache *);

		for (i = 0; i < nbc; i++) {
			/* BUG: should check intervening caches too */
			if (!bc[i]->c_wthru) {
				shade_error("%s: %c%d not write thru", save,
					    bc[i]->c_idu, bc[i]->c_level);
				return(FALSE);
			}
			if (bc[i]->c_bbsize < 0) {
				shade_error("%s: can't include null cache (%c%d)",
					    save, bc[i]->c_idu, bc[i]->c_level);
				return(FALSE);
			}
				    
			c->c_bkinv[i] = bc[i];
			c->c_bkinv[i]->c_isbkinv = 1;
		}
		c->c_bkinv[i] = 0;
	}

	if (wi)
		if (c->c_wthru && !c->c_walloc && c->c_nsets == 1 && 
		    nlevels == 0)
		{
			c->c_winv = 1;
		}
		else {
			shade_error("%s: wi not allowed for this cache", save);
			return(FALSE);
		}

	c->c_cwb = cwb;

	return(TRUE);
}


static void newcachesets(
	Cache * c)
{
	int	s;
	int	alloc_failed = FALSE;
	switch (c->c_replace) {
	case C_LRU:
		c->c_lru = 1 + alloc (c->c_nsets + 1, spix_uint32_t *);
		for (s = -1; s < c->c_nsets; s++) {
			c->c_lru[s] = allocz (c->c_nlines, spix_uint32_t);
		}
		break;
	case C_RANDOM:
		break;
	case C_VIKING:
		c->c_mru = allocz (c->c_nlines, mask_t);
		break;
	case C_PLRU:
		if (c->c_nsets != 4) {
			shade_error ("Pseudo-LRU only supports 4-way");
			exit(1);
		}
		c->c_plru = allocz (c->c_nlines, char);
		break;
	case C_LFSR:
		if (c->c_nsets != 4) {
			shade_error ("LFSR only supports 4-way");
			exit(1);
		}
		break;
	case C_NVLFSR:
		if (c->c_nsets != 4) {
			shade_error ("NVLFSR only supports 4-way");
			exit(1);
		}
		c->c_nxtvctm = allocz (c->c_nlines, char);
		break;
	case C_NVCTR:
		if (ilog2(c->c_nsets) <= 1) {
			shade_error ("NVCTR only supports N-way where N is more more than 1 and a power of 2");
			exit(1);
		}
		c->c_nxtvctm = allocz (c->c_nlines, char);
		break;
	}
}


static int replacecacheset(
	Cache * c,
	addr_t	X)
{
	int	s, s2, s3;
	mask_t	mru;

	switch (c->c_replace) {
	case C_LRU:
		for (s2 = 0, s = 1; s < c->c_nsets; s++)
			if (c->c_lru[s][X] < c->c_lru[s2][X])
				s2 = s;
		return (s2);
	case C_RANDOM:
		s = rand() % c->c_nsets;
		return (s);
	case C_VIKING:
		mru = c->c_mru[X];
		for (s = c->c_nsets-1; s > 0 && (mru & (1 << s)); s--);
		return (s);
	case C_PLRU:
		s2 = c->c_plru[X];
		if (s2 & 4)
		{
			/* Set 0 or 1 was accessed most recently.
			 * Look at LSB to choose set 2 or 3. */
			if (s2 & 1)
				return 2;
			else return 3;
		}
		else
		{
			if (s2 & 2)
				return 1;
			else return 0;
		}
	case C_LFSR:
				/* Invert odd bits: */
				/* lfsr_out */ s2 = c->c_seed ^ /*'b10101*/ 21;
	/* rand_in */  s = s2 & 3; /* take low 2 bits */
	/* lfsr_out[0] */ s3 = s2 & 1;
	/* lfsr_in */  s2 = (s3<<4)
			       | ((((s3<<3)|(s3<<2)|(s3<<1)|s3)
				   & /*'b1011*/11)
				  ^ ((s2 & 30)>>1));
	/* Invert odd bits: */
	c->c_seed = s2 ^ /*'b10101*/ 21;
	return (s);
	case C_NVLFSR:
		s = c->c_nxtvctm[X];
				/* Invert odd bits: */
				/* lfsr_out */ s2 = c->c_seed ^ /*'b10101*/ 21;
				/* rand_in */  c->c_nxtvctm[X] = s2 & 3; /* take low 2 bits */
				/* lfsr_out[0] */ s3 = s2 & 1;
				/* lfsr_in */  s2 = (s3<<4)
						       | ((((s3<<3)|(s3<<2)|(s3<<1)|s3)
							   & /*'b1011*/11)
							  ^ ((s2 & 30)>>1));
				/* Invert odd bits: */
		c->c_seed = s2 ^ /*'b10101*/ 21;
		return (s);
	case C_NVCTR:
		s = c->c_nxtvctm[X];
		c->c_nxtvctm[X] = c->c_seed;
		c->c_seed = (c->c_nsets - 1) & (1 + c->c_seed);
		return (s);
	}
	return (0);
}


static void resetcache(
	Cache * c)
{
	invalidatecache (c);

	if (c->c_nsets > 1)
		resetcachesets (c);
}


static void resetcaches(void)
{
	int	i;
	int	j;
	
	for (j=0; j<cpus_used; j++)
		for (i = 0; i < nlevels; i++) {
			resetcache (cpu_caches[j]->icaches + i);
			resetcache (cpu_caches[j]->dcaches + i);	/* overkill if unified */
		}
}


static void resetcachesets(
	Cache * c)
{
	addr_t	X;
	int	s;

	switch (c->c_replace) {
	case C_LRU:
		for (s = -1; s < c->c_nsets; s++)
			for (X = 0; X < c->c_nlines; X++)
				c->c_lru[s][X] = 0;
		break;
	case C_RANDOM:
		srand(1);
		break;
	case C_VIKING:
		for (X = 0; X < c->c_nlines; X++)
			c->c_mru[X] = 0;
		break;
	case C_PLRU:
		for (X = 0; X < c->c_nlines; X++)
			c->c_plru[X] = 0;
		break;
	case C_LFSR:
		c->c_seed = 0;
		break;
	case C_NVLFSR:
		for (X = 0; X < c->c_nlines; X++)
			c->c_nxtvctm[X] = 0;
		c->c_seed = 0;
		break;
	case C_NVCTR:
		for (X = 0; X < c->c_nlines; X++)
			c->c_nxtvctm[X] = 0;
		c->c_seed = 0;
		break;
	}
}


/*
**  Clear out all the statistics gathered so far.
*/
static void clear_cachestats(void)
{
	int	i;
	int	j;
	
	for (j=0; j<cpus_used; j++)
		for (i = 0;  i < nlevels;  i++) {
			clear_cachestat(&cpu_caches[j]->icaches[i]);
			clear_cachestat(&cpu_caches[j]->dcaches[i]);
		}
	
	nina = 0;
}


static void clear_cachestat(
	Cache * pc)
{
	int	sz;


	for (sz = 0;  sz <= MAXSIZE;  sz++) {
		pc->c_nref[sz][0] = 0;
		pc->c_nref[sz][1] = 0;
		pc->c_nmiss[sz][0] = 0;
		pc->c_nmiss[sz][1] = 0;
	}

	pc->c_ndwback = 0;
	pc->c_ncwback = 0;
	pc->c_nrw = 0;
	pc->c_nbkinving = 0;
	pc->c_nbkinved = 0;
	pc->c_nwinv = 0;
}


static void cachestats(
	FILE *	fp,
	Cache * icaches,
	Cache * dcaches,
	count_t ni)
{
	Cache	*ic, *dc;

	fprintf (fp, "\n\n");

	cachestatsc (fp, (Cache *) 0);

	for (ic=icaches, dc=dcaches; ic&&dc; ic=ic->c_next, dc=dc->c_next) {
		cachestatsc (fp, ic);
		if (dc->c_idu == 'D')
			cachestatsc (fp, dc);
	}

	if (ni <= 0)
		return;

	for (ic=icaches, dc=dcaches; ic&&dc; ic=ic->c_next, dc=dc->c_next)
		cachestatsl (fp, ic, dc, ni);
}


static void cachestatsc(
	FILE *	fp,
	Cache * c)
{
	Cache	**cc;
	char	wp[32];

	if (!c) {
		fprintf (fp, "%-5s%8s%8s%8s%8s	%-7s  %-18s  %s\n\n",
			 "cache", "size", "block", "subblk", "assoc",
			 "replace", "write policy", "includes");
		return;
	}

	fprintf (fp, "%c%-4d", c->c_idu, c->c_level);

	if (c->c_ssize < 0) {
		fprintf (fp, "%8s\n", "0   ");
		return;
	}

	fprintf (fp, "%8s", putsize ((addr_t) c->c_nsets << c->c_ssize));

	fprintf (fp, "%6u B", 1 << c->c_bsize);

	if (c->c_bsize != c->c_bbsize)
		fprintf (fp, "%6u B", 1 << c->c_bbsize);
	else	fprintf (fp, "%8s", "N/A");

	if (c->c_nsets <= 1)
		fprintf (fp, "%8s", "direct");
	else	fprintf (fp, "%4d-way", c->c_nsets);

	if (c->c_nsets > 1)
		if (c->c_replace == C_LRU)
			fprintf (fp, "	%-7s", "lru");
		else if (c->c_replace == C_RANDOM)
			fprintf (fp, "	%-7s", "random");
		else if (c->c_replace == C_VIKING)
			fprintf (fp, "	%-7s", "viking");
		else if (c->c_replace == C_PLRU)
			fprintf (fp, "	%-7s", "p-lru");
		else if (c->c_replace == C_LFSR)
			fprintf (fp, "	%-7s", "lfsr");
		else if (c->c_replace == C_NVLFSR)
			fprintf (fp, "	%-7s", "nvlfsr");
		else if (c->c_replace == C_NVCTR)
			fprintf (fp, "	%-7s", "nvctr");
		else	fprintf (fp, "	%-7s", "N/A");
	else	fprintf (fp, "	%-7s", "N/A");

	sprintf (wp, "%s,%salloc%s%s",
		 c->c_wthru ? "thru" : "back",
		 c->c_walloc ? "" : "no",
		 c->c_wthru && !c->c_walloc ? (c->c_winv ? ",inv" : ",noinv") : "",
		 c->c_cwb ? ",clwb" : "");
	fprintf (fp, "  %-18s ", wp);

	if (cc = c->c_bkinv)
		for (; *cc; cc++)
			fprintf (fp, "%c%d%s", (*cc)->c_idu, (*cc)->c_level,
				 cc[1] ? "," : "\n");
	else	fprintf (fp, "N/A\n");
}


static void cachestatsl(
	FILE *	fp,
	Cache * ic,
	Cache * dc,
	count_t ni)
{
	count_t r, ref, iref, dref, idrwref[2][2], m, miss, idrwmiss[2][2];
	int	id, rw, sz;

	/*
	 * for null caches, all references miss.
	 */
	if (ic->c_bbsize < 0)
		for (sz = 0; sz <= MAXSIZE; sz++)
			for (rw = 0; rw < 2; rw++)
				ic->c_nmiss[sz][rw] = ic->c_nref[sz][rw];
	if (dc->c_bbsize < 0)
		for (sz = 0; sz <= MAXSIZE; sz++)
			for (rw = 0; rw < 2; rw++)
				dc->c_nmiss[sz][rw] = dc->c_nref[sz][rw];

	/*
	 *  Get summary values for reference counters
	 */
	iref = dref = ref = miss = 0;
	for (id = 0; id < 2; id++)
		for (rw = 0; rw < 2; rw++)
			idrwref[id][rw] = idrwmiss[id][rw] = 0;

	for (sz = 0; sz <= MAXSIZE; sz++)
		for (rw = 0; rw < 2; rw++) {
			idrwref[0][rw] += ic->c_nref[sz][rw];
			idrwref[1][rw] += dc->c_nref[sz][rw];
			idrwmiss[0][rw] += ic->c_nmiss[sz][rw];
			idrwmiss[1][rw] += dc->c_nmiss[sz][rw];
		}

	for (rw = 0; rw < 2; rw++) {
		iref += idrwref[0][rw];
		dref += idrwref[1][rw];
		for (id = 0; id < 2; id++)
			miss += idrwmiss[id][rw];
	}
	ref = iref + dref;

	fprintf (fp, "\n\n%s Level:\n", ordinal (ic->c_level));
	fprintf (fp, "%15s%11s%11s\n\n",
		 "#", "%instrs", "%I+Drefs");

#define doita(x,y)	cachestatsla (fp, (x), ni, ref), \
  fprintf (fp, "	%s", (y))
#define doitb(a,b,c)	fprintf (fp, "	(%.3f%%%s)\n", \
				 (b) ? percent ((a), (b)) : 100., (c))

	doita (ref, "I+D references\n");
	doita (miss, "I+D misses\n");
	fprintf (fp, "\n");
	
	doita (iref, "I references\n");
	doita (m = idrwmiss[0][0] + idrwmiss[0][1], "I misses");
	doitb (m, iref, "");
	fprintf (fp, "\n");

	doita (dref, "D references\n");
	doita (m = idrwmiss[1][0] + idrwmiss[1][1], "D misses");
	doitb (m, dref, "");
	fprintf (fp, "\n");

	doita (r = idrwref[1][0], "D reads");
	doitb (r, dref, " D refs");
	doita (m = idrwmiss[1][0], "D read misses");
	doitb (m, r, "");
	fprintf (fp, "\n");

	doita (r = idrwref[1][1], "D writes");
	doitb (r, dref, " D refs");
	doita (m = idrwmiss[1][1], "D write misses");
	doitb (m, r, "");
	if (!dc->c_wthru)
		doita (dc->c_nrw, "D read mod writes\n");
	if (dc->c_winv) {
		doita (dc->c_nwinv,"D write invalidates");
		doitb (dc->c_nwinv, idrwref[1][1], " D writes");
	}
	fprintf (fp, "\n");

	if (!ic->c_wthru)    /* I miss causes data write back in I+D cache */
		doita (ic->c_ndwback, "dirty write backs due to I misses\n");
	if (!dc->c_wthru)
		doita (dc->c_ndwback, "dirty write backs due to D misses\n");
	if (ic->c_cwb)
		doita (ic->c_ncwback, "clean write backs due to I misses\n");
	if (dc->c_cwb)
		doita (dc->c_ncwback, "clean write backs due to D misses\n");
	if (!ic->c_wthru || !dc->c_wthru)
		doita (ic->c_ndwback + dc->c_ndwback, "dirty write backs\n");
	if (ic->c_cwb || dc->c_cwb)
		doita (ic->c_ncwback + dc->c_ncwback, "clean write backs\n");
	if (!ic->c_wthru || !dc->c_wthru || ic->c_cwb || dc->c_cwb)
		doita (ic->c_ndwback + dc->c_ndwback + ic->c_ncwback + dc->c_ncwback, "write backs\n\n");

	if (ic->c_isbkinv)
		doita (ic->c_nbkinved, "I back invalidated subblocks\n");
	if (dc->c_isbkinv)
		doita (dc->c_nbkinved, "D back invalidated subblocks\n");
	if (ic->c_bkinv)
		doita (ic->c_nbkinving, "I victimized subblocks\n");
	if (dc->c_bkinv)
		doita (dc->c_nbkinving, "D victimized subblocks\n");
	if (ic->c_bkinv || ic->c_isbkinv ||
	    dc->c_bkinv || dc->c_isbkinv)
		fprintf (fp, "\n");


	for (sz = 0; sz <= MAXSIZE; sz++)
		for (rw = 0; rw < 2; rw++) {
			static char *rws[2] = {"read", "write"};

			r = dc->c_nref[sz][rw];
			m = dc->c_nmiss[sz][rw];

			/*
			 * for uniform output, always print all counts
			 * for 1-8 byte references at 1st level.
			 * otherwise only print nonzero counts.
			 */
			if (!r && !m && (sz > 3 || dc->c_level > 1))
				continue;

			cachestatsla (fp, r, ni, ref);
			fprintf (fp, "	%s D %ss  (%.3f%% D %ss)\n", 
				 putsize ((addr_t) 1 << sz), rws[rw],
				 percent (r, idrwref[1][rw]), rws[rw]);

			cachestatsla (fp, m, ni, ref);
			fprintf (fp, "	%s D %s misses	(%.3f%%)\n", 
				 putsize ((addr_t) 1 << sz), rws[rw],
				 r ? percent (m, r) : (m ? 100. : 0.));
		}

	fprintf (fp, "\n");
}


static void cachestatsla(
	FILE *	fp,
	count_t x,
	count_t ni,
	count_t ref)
{
	fprintf (fp, "%15lld%10.4f%%", x, percent (x, ni));
	if (ref) 
		fprintf (fp, "%10.4f%%", percent (x, ref)); 
	else	
		fprintf (fp, "%11s", "");
}

static void 
cachevalidstats (FILE	* fp, Cache	*c)
{
	int	n, nsb, nvalid = 0, ndirty = 0, s, X, sb;
	mask_t	Y;

	if (c->c_nsets <= 0 || c->c_nlines <= 0)
		return;

	nsb = 1 << (c->c_bsize - c->c_bbsize);

	for (s = 0; s < c->c_nsets; s++)
		for (X = 0; X < c->c_nlines; X++)
			for (sb = 0, Y = 1; sb < nsb; sb++, Y <<= 1) {
				if (c->c_valid[s][X] & Y)
					nvalid++;
				if (c->c_valid[s][X] & c->c_dirty[s][X] & Y)
					ndirty++;
			}

	n = c->c_nsets * c->c_nlines * nsb;

	fprintf (fp, "%15d  %c%d lines valid (%.4f%% lines)\n",
		 nvalid, c->c_idu, c->c_level, percent (nvalid, n));
	fprintf (fp, "%15d  %c%d lines dirty (%.4f%% lines)\n",
		 ndirty, c->c_idu, c->c_level, percent (ndirty, n));
	fprintf (fp, "%15d  %c%d lines clean (%.4f%% lines)\n",
		 nvalid-ndirty, c->c_idu, c->c_level, percent (nvalid-ndirty, n));
}

static void cacheusestats(
	FILE *	fp,
	Cache * c)
{
	int	n, nused = 0, s, X;

	if (c->c_nsets <= 0 || c->c_nlines <= 0 || !c->c_used)
		return;

	for (s = 0; s < c->c_nsets; s++)
		for (X = 0; X < c->c_nlines; X++)
			if (c->c_used[s][X])
				nused++;

	n = c->c_nsets * c->c_nlines;

	fprintf (fp, "%15d  %c%d blocks used (%.4f%% blocks)\n",
		 nused, c->c_idu, c->c_level, percent (nused, n));
	fprintf (fp, "%15d  %c%d blocks unused (%.4f%% blocks)\n",
		 n-nused, c->c_idu, c->c_level, percent (n-nused, n));
	cachevalidstats(fp,c);
}


static void *domalloc(unsigned int	n)
{
	void *	x;

	x = shade_malloc(n);
	if (!x && n) {
		shade_error("can't malloc %u bytes", n);
		exit(1);
	}
	return (x);
}


static void *docalloc(unsigned int	n)
{
	void *	x;

	x = shade_malloc(n);
	if (!x && n) {
		shade_error("can't malloc %u bytes", n);
		exit(1);
	}
	memset(x, 0, n);
	return (x);
}


static addr_t getsize(
	char *	s,
	char ** ss)
{
	addr_t	u;

	u = strtol (s, &s, 10);

	for (; *s; s++)
		if (*s == 'K')
			u <<= 10;
		else if (*s == 'M')
			u <<= 20;
		else if (*s == 'G')
			u <<= 30;
		else	break;

	if (ss)
		*ss = s;
	return (u);
}


static int ilog2(
	addr_t	u)
{
	int	i;

	if (u == 0)
		return (-1);

	for (i = 0; (u & 1) == 0; i++, u >>= 1);

	return (u == 1 ? i : -1);
}


static void init_memsize(void)
{
	int	ih, sz;
	size_t	sz_mem;

	for (ih = 0; ih < SPIX_SPARC_NIOP; ih++) {
		sz_mem = spix_sparc_iop_memsize(ih);
		for (sz = 0; sz_mem > (1<<sz) && sz < MAXSIZE; sz++);
		ihash_memsize[ih] = sz;
	}
}


static char *ordinal(
	int	card)
{
	static char	buf[32], *suf[] = { "th", "st", "nd", "rd" };
	int		n = card % 10;

	if (card >= 11 && card <= 13 || n > 3)
		n = 0;
	sprintf (buf, "%d%s", card, suf[n]);
	return (buf);
}


static char *putsize(
	addr_t	u)
{
	static char	buf[16];
	char		*s = "B KBMBGBTB";

	for (; s[2] && (u & 0x3ff) == 0; s += 2, u >>= 10);
	sprintf (buf, "%u %2.2s",(unsigned)u, s);
	return (buf);
}

/* Selects the cpu for given thread ID */
static CpuCache* 
get_cpu(int th_id) {
	int i;
	if (cpus_used <= th_id) {

		if (cpus_allocated <= th_id) {
			if (!reallocate_caches(th_id)) {
				shade_error("Unable to reallocate caches for new thread, th_id=%d\n",th_id);
				exit(1);
			}
		}
		/* we have space, initialize new cpu(s) */
		for (i=cpus_used; i <= th_id; i++) {
			cpu_caches[i] = new_cpu(cpu_caches[0]);	     
			if (!cpu_caches[i]) {
				shade_error("Unable to allocate cache for new thread, "
					    "th_id=%d, used=%d, allocated=%d, nlevels=%d\n",
					    th_id, i, cpus_allocated, nlevels);
				exit(1);
			}
		}
		cpus_used = i;
	} 
	return cpu_caches[th_id];
}

/* Reallocates array of cpu caches */
static int
reallocate_caches(int th_id)
{
	CpuCache** new_line;
	int	   i;

	new_line = allocz(th_id+THDS_CHUNK, CpuCache*);
	if (!new_line)
		return FALSE;

	/* copy pointers */
	for (i=0; i< cpus_allocated; i++)
		new_line[i] = cpu_caches[i];
	cpus_allocated = th_id+THDS_CHUNK;

	/* free old array and store new array pointer */
	free(cpu_caches);
	cpu_caches = new_line;
	return TRUE;
}


/* Creates new icaches/dcaches for the cpu being added */
static CpuCache*
new_cpu(CpuCache* ex) {
	CpuCache* new_cache;
	Cache**	  pbk;
	Cache*	  c;
	int	  i, level;
	int	  is_d;

	new_cache = (CpuCache*)alloc(1,CpuCache);

	/* copy plain old data information */
	*new_cache = *ex;


#if 1
	/* restore proper chaining */
	for (i = 0; i < nlevels-1; i++) {
		new_cache->icaches[i].c_next = new_cache->icaches + i + 1;
		new_cache->dcaches[i].c_next = new_cache->dcaches + i + 1;
	}
#endif

  
	for (level = 0; level < nlevels*2; level++) {
		is_d = (level >= nlevels);
		/* copy all c & d caches info */

		/* is it a dcache ? */
		c = (is_d ? new_cache->dcaches+(level-nlevels) :
		     new_cache->icaches+level);

		if ((c->c_idu=='u' || c->c_idu=='U')&& is_d) {
				/* for unified cache, 'd' is a shallow copy of 
				   corresponding 'i' */
			*c = *(new_cache->icaches+(level-nlevels));
			continue;
		}

		/* reallocate basic arrays */
		c->c_tags = alloc (c->c_nsets, addr_t *);
		c->c_valid = alloc (c->c_nsets, mask_t *);
		if (!c->c_wthru)
			c->c_dirty = alloc (c->c_nsets, mask_t *);
		c->c_used = alloc (c->c_nsets, char *);

		/* realloc replacement, if any */
		newcachesets(c);

		for (i = 0; i < c->c_nsets; i++) {
			c->c_tags[i] = allocz (c->c_nlines, addr_t);
			c->c_valid[i] = allocz (c->c_nlines, mask_t);
			if (!c->c_wthru)
				c->c_dirty[i] = allocz (c->c_nlines, mask_t);
			c->c_used[i] = allocz (c->c_nlines, char);
		}

		/* count and copy back invalidate info */
		for (i=0,pbk = c->c_bkinv; pbk && *pbk; pbk++, i++);
    
		if (i) {
			Cache**	  pc;

			pc = (Cache**)allocz(i+1,Cache*);
			for(--i;i>=0;--i)
				pc[i] = new_cache->icaches+(c->c_bkinv[i]-ex->icaches);
				/* reassign back invalidate info to new array */
			c->c_bkinv = pc;
		}

		/* done copying the cache, reset it */
		resetcache(c);
		/* clear statistics */
		clear_cachestat(c);	
	}

#if 1
	/* restore proper chaining */
	for (i = 0; i < nlevels-1; i++) {
		new_cache->icaches[i].c_next = new_cache->icaches + i + 1;
		new_cache->dcaches[i].c_next = new_cache->dcaches + i + 1;
	}
#endif

	return new_cache;
}


/* binds all lwp's to single processor.
   Necessary for running cachesim5 in single-cpu mode
   on a multiprocessor */
static void             
bind_lwps()
{
	processorid_t i;
	int status;
	int n = sysconf(_SC_NPROCESSORS_ONLN);
	if (n==1) /* nothing to do */
		return;
	for (i=0; ; i++) {
		status = p_online(i, P_STATUS);
		if ((status==-1)&&(errno==EINVAL))
			continue;
		else
			break;
	}
	/* i is valid processor id, bind the process lpws to it */
	status = processor_bind(P_PID, P_MYID, i, NULL);
	if (status == -1)
		shade_error("Was not able to bind Shade process to processor %d, errno == %d\n",
			    i, errno);	
}
