/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  ifreq.c	- Shade analyzer to count opcode frequencies
**
**	This file defines a Shade analyzer that counts the number of
**	dynamic instructions executed by the target application, sorted by
**	opcode.
*/

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <spix_sparc.h>
#include <shade.h>
#include <shade_anal.h>
#include "getopt.h"
#include <sun_prefetch.h>

/*
 * Define a standard trace record.
 */
struct shade_trace_s {SHADE_TRACE};

const char	shadeuser_analversion[] = "ifreq " "v1.0.1";


/*
 * Per-opcode instruction counts for executed and annulled instructions.
 */
static unsigned long long	ihashExec[SPIX_SPARC_NIOP];
static unsigned long long	ihashAnnul[SPIX_SPARC_NIOP];
static uint64_t			total_instructions = 0;

/* Profile collection variables*/
static FILE *           	fpOutput = stdout;
static int profile =0;
static uint64_t 	*exec_counts[32768];

/*
 * The process ID of the first Shade process.  Trace output from all other
 * processes will be labeled as coming from a child process.
 */
static pid_t	        pidFirst;

/*Options for profile collection*/
enum {
      IFREQ_OPT_PROF,
};

const shade_options_t shadeuser_opts[]={
    {"-prof", "Writes profile information to file", NULL, STR_PARAM},
    {0, "", NULL, NO_PARAM}
};

static int 	cmpops(const void *, const void *);

extern void ifreq_profile_init();
extern void ifreq_profile_ref();
extern void ifreq_profile_report();
extern void ifreq_profile_clear();

/*
**  Perform initializations specific to the 'ifreq' analyzer.  Return zero
**  on success, non-zero on failure.
*/
int shadeuser_initialize(
	int     	argc,
	char ** 	argv,
	char ** 	envp)
{
	shade_iset_t *	piset;
	shade_tset_t *	ptset;
	spix_bool_t	ok;
	int     	i, j;
	int                 ifreq_optnum;
	shade_option_val_t  ifreq_opt_value;
    	/*name of profile file*/
	const char* profile_out=0;
	int index;
	char * filename;

        /*
         * actually this is not records (should me multiplied by 1024)
         */



	ok = TRUE;

	for (i = j = 1;  i < argc;  i++) {
	  ifreq_optnum = shade_getoptv(argc, argv, &i, &j, shadeuser_opts, &ifreq_opt_value);
          switch(ifreq_optnum)
          {
            case IFREQ_OPT_PROF:
	        profile_out = ifreq_opt_value.sval;
		profile=1;
                break;
    	    default:
            	shade_error("Unknown option '%s'.", argv[i]);
	        ok = FALSE;
	        break;
	  }
	}


	if (!ok) {
		shade_usage();
		return(1);
	}

	/*
	 * Record the initial process ID.
	 */
	pidFirst = getpid();

	/*
	 * We want to be notified of fork() operations, so we know if the
	 * application spawns a child process.
	 */
	shade_setopt(SHADE_OPT_FORKNOTIFY);

	/*
	 * Set up the trace control parameters to trace all instructions,
	 * both executed and annulled.
	 */
	shade_trsize(sizeof(shade_trace_t));
	piset = shade_iset_newclass(SHADE_ICLASS_ANY, -1);
  	if (profile)
	{
	 ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_ANNULLED, SHADE_TRCTL_PC, -1);
	}
	else
	{
	 ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_ANNULLED, -1);
	}
	shade_trctl(piset, SHADE_TRI_ISSUED, ptset);

	/*Open profile file for output*/
	if (profile_out) {
		index = 1;
		filename = (char*)malloc(strlen(profile_out)+40);
		sprintf(filename,"%s",profile_out); /*Make name*/

		fpOutput = fopen(filename, "r");
		if (fpOutput!=NULL)
		{
			while (fpOutput!=NULL)
			{			
				fclose(fpOutput); /*Close opened file*/						
				sprintf(filename,"%s.%d",profile_out,index); /*Make name*/
				fpOutput = fopen(filename, "r");
				fprintf(stderr,"%s\n",filename);
				index++;
			}
		}

		/*got unique name*/
		fpOutput = fopen(filename, "w");
		if (!fpOutput) 
		{
		    fprintf(stderr, "Unable to open '%s'\n", filename);
		    free(filename);
		    exit(1);
		}
		free(filename);

		ifreq_profile_init();
	}

	return(0);
}


/*
**  Run and trace an application that is loaded into Shade.
*/
int shadeuser_analyze(void)
{
	shade_trace_t   	atr[5*1024], *patr, *patr_last;

	shade_status_t  	status;
	static pid_t           	pid_parent;
	pid_t           	pid;
	int             	ntr;
	int              	i, *ppi;
	int             	val;
	uint64_t pc;

	status = shade_appstatus(&val);

	switch (status) {
	case SHADE_STATUS_LOADED:

	  /*
	   * Remember the parent process ID so we can tell if it forks.
	   */
	  pid_parent = getpid();
	  
	  break;

	case SHADE_STATUS_FORKED:
	  /*
	   * If the application forks, zero the counts in the child.
	   * The child shouldn't inherit execution counts from the
	   * parent.
	   */
	  pid = getpid();
	  if (pid != pid_parent) {
	    pid_parent = pid;
	    memset(ihashExec, 0, sizeof(ihashExec));
	    memset(ihashAnnul, 0, sizeof(ihashAnnul));
	    /*Clear the profile array*/
	    ifreq_profile_init();
	  }
	  
	  break; 
	}

	/*
	 * Count all instructions of each opcode.
	 */


	ntr = shade_run(atr, sizeof(atr)/sizeof(*atr));

        if (ntr == 0)
          return 0;


	for (i = 0;  i < ntr;  i++) {
	  sparc_prefetch_read_many(&atr[i+8].tr_pc);
	  if (atr[i].tr_annulled)
	    ihashAnnul[atr[i].tr_ih]++;
          else
	    ihashExec[atr[i].tr_ih]++;

	  /*Handle profile data*/
	  if (profile) {
	    pc = (uint64_t) atr[i+2].tr_pc;
	    ifreq_profile_ref(pc);
	  }
	}

	return(0);
}

void
shadeuser_report(int reason, void* ptr) {
	unsigned long long	sum;
	unsigned long long	sumann;
	unsigned long long	cnt;
	unsigned long long	cntann;
	unsigned          	ihs[SPIX_SPARC_NIOP];
	unsigned          	ih;
	unsigned          	i;
	pid_t           	pid;
  
	switch (reason) {
	case SHADEUSER_REPORT_TERMINATE:
		/* no info is consolidated for applications */
		break;
	  
	case SHADEUSER_REPORT_SIGNAL:		
	case SHADEUSER_REPORT_APP_DONE:		
		/*
		 * Sort the opcode frequency table.
		 */
		for (sum = 0, sumann = 0, i = 0;  i < SPIX_SPARC_NIOP;  i++) {
			sum += ihashExec[i];
			sumann += ihashAnnul[i];
			ihs[i] = i;
		}
		
		qsort(ihs, SPIX_SPARC_NIOP, sizeof(ihs[0]), cmpops);
		
		/*
		 * If this is a child process, indicate that in the banner.
		 */
		pid = getpid();
		if (pid != pidFirst)
			fprintf(shade_fp, "\n\nOPCODES (child process %d):\n\n", pid);
		else
			fprintf(shade_fp, "\n\nOPCODES:\n\n");
		
		/*
		 * Print the entries.
		 */
		fprintf(shade_fp, "%-*s%15s%11s%14s\n\n", 
			spix_sparc_iop_Lname, "opcode", "#exec", "%exec", "#annulled");
		
		for (i = 0;  i < SPIX_SPARC_NIOP;  i++) {
			ih = ihs[i];
			cnt = ihashExec[ih];
			cntann = ihashAnnul[ih];
			if (cnt || cntann) {
				fprintf(shade_fp, "%-*s%15llu%10.4f%%%14llu\n",
					spix_sparc_iop_Lname, spix_sparc_iop_name(ih),
					cnt, 100. * cnt / (double)sum,
					cntann);
			}
		}
		
		fprintf(shade_fp, "%-12s%15llu%10.4f%%%14llu\n\n\n", 
			"TOTAL", sum, 100.0, sumann);

		/*Output profile*/
		if (profile) {
		  if (pid != pidFirst)
		    {
		      fprintf(fpOutput, "//ifreq profile (child process %d) [address] [count]:\n",
			      pid);
		    }
		  else
		    fprintf(fpOutput, "//ifreq profile [address] [count]:\n");

		  ifreq_profile_report(fpOutput);
		}
		break;
	}
}

/*
**  Perform termination activities after all applications have been executed.
*/

int shadeuser_terminate(
	int	ret)
{
	int	val;

	/*
	 * Exit with the termination status from the executed application.
	 */
	switch (shade_appstatus(&val))
	{
	case SHADE_STATUS_EXITED:
		return(val);

	case SHADE_STATUS_SIGNALED:
		return(128 + val);

	case SHADE_STATUS_NOAPP:
		return(0);

	case SHADE_STATUS_LOADED:
	case SHADE_STATUS_RUNNING:
	default:
		assert(0);
	}

}

/*
**  Qsort comparison function to order the ihashExec[] array by increasing
**  execution counts.
*/
static int cmpops(
	const void *	pvidx1,
	const void *	pvidx2)
{
	const unsigned *	pidx1 = pvidx1;
	const unsigned *	pidx2 = pvidx2;
	unsigned        	idx1;
	unsigned        	idx2;
	unsigned long long	cnt1;
	unsigned long long	cnt2;


	idx1 = *pidx1;
	idx2 = *pidx2;
	cnt1 = ihashExec[idx1];
	cnt2 = ihashExec[idx2];

	if (cnt1 < cnt2)
		return(1);
	if (cnt1 > cnt2)
		return(-1);

	return(strcmp(spix_sparc_iop_name(idx1), spix_sparc_iop_name(idx2)));
}


/*
**  Print a usage message for the command line options that are specific
**  to the 'ifreq' analyzer.
*/
void shadeuser_analusage(void)
{
    shade_print_opt_info(shadeuser_opts, 0);
}


struct ifreq_profile_elem {
  uint64_t pc;
  int64_t count;

  struct ifreq_profile_elem * next;
};

static struct ifreq_profile_elem ** table;
static const int tablesize = 1<<20;

void ifreq_profile_init()
{
  int idx;

  table = (struct ifreq_profile_elem **) malloc(tablesize * sizeof(struct ifreq_profile_elem *));
  for (idx=0; idx<tablesize; idx++) {
    table[idx] = NULL;
  }
} /* void ifreq_profile_init() */



void ifreq_profile_ref(uint64_t pc)
{
  int idx = (pc >> 2) & (tablesize-1);
  struct ifreq_profile_elem * elem;
  for (elem=table[idx]; elem != NULL; elem=elem->next) {
    if (elem->pc == pc) {
      elem->count++;
      break;
    }
  }
  if (elem==NULL) {
    elem = malloc(sizeof(struct ifreq_profile_elem));
    elem->pc = pc;
    elem->count=1;
    elem->next = table[idx];
    table[idx] = elem;
  }
} /* void ifreq_profile_ref(uint64_t pc) */

void ifreq_profile_report(FILE * fp)
{
  int idx;
  for (idx=0; idx<tablesize; idx++) {
    struct ifreq_profile_elem * elem;
    for (elem = table[idx]; elem != NULL; elem=elem->next) {
      fprintf(fp, "%#15llx \t%15llu\n", elem->pc, elem->count);
    }
  }
} /* void ifreq_profile_report(FILE * fpOutput) */


void ifreq_profile_clear()
{
  int idx;
  for (idx=0; idx<tablesize; idx++) {
    struct ifreq_profile_elem * elem = table[idx];
    while(elem != NULL) {
      table[idx] = elem->next;
      free(elem);
    }
  }

  free(table);
}
