/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  window.c	- Shade analyzer to measure register window statistics
**
**	This Shade analyzer measures various register window statistics
**	including the number of overflows and underflows for various
**	register window sizes, the maximum save depth, and the maximum
**	number of consecutives saves and restores.
*/

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <sys/trap.h>
#include <sys/syscall.h>
#include <spix_sparc.h>
#include <shade_sparcv9.h>
#include <shade_anal.h>
#include "getopt.h"

/*
 * Define a standard SPARC V9 trace record.
 */
struct shade_trace_s {SHADE_SPARCV9_TRACE};

const char	shadeuser_analversion[] = "window " "v1.0.1";


#define percent(a,b)	(100. * (a) / (b))


/*
 * Some special "opcode" values we use to identify flushwindow trap
 * instructions and setcontext system call trap instructions.  (See
 * pre_ticc().)
 */
#define TICC_FLUSHWINDOWS	SPIX_SPARC_NIOP + 1
#define TICC_SETCONTEXT 	SPIX_SPARC_NIOP + 2


/*
 * Some interesting register numbers on SPARC.
 */
#define REG_G1	1
#define REG_O0	8
#define REG_FP	30


#define NWCS		128
#define MAXDEPTH	1024
#define MAXRUN		256
#define MAXTHDS 	256


typedef struct {
	int		wc_size;        	/* # windows in register file */
	unsigned	wc_depth[MAXTHDS];	/* each thread's position */
	/*     [1,wc_size] */
	unsigned	wc_nover;       	/* # overflows */
	unsigned	wc_nunder;      	/* # underflows */
	unsigned	wc_nflush;      	/* # flushes */
} Wcntr;



static Wcntr	wcs[NWCS];	/* stats per register file config */

static double	sumdepth,	/* sum of "depth" values */
	sumrestrun,	/* sum of restore run lengths */
	sumsaverun;	/* sum of save run lengths */

static unsigned depthcnts[MAXDEPTH],	/* counts per save depth */
	maxdepth,		/* max save depth */
	maxrestrun,		/* max # restores with no saves */
	maxsaverun,		/* max # saves with no restores */
	nflushwindows,		/* # flush window traps (ta 3) */
	nrestores,		/* # restore instructions exec'd */
	nrestruns,		/* # restore instruction runs */
	nsaves,			/* # save instructions exec'd */
	nsaveruns,		/* # save instruction runs */
	nsetcontexts,		/* # setcontext system calls */
	restruncnts[MAXRUN],	/* counts per restore run length */
	saveruncnts[MAXRUN],	/* counts per save run length */
	setcontexts[MAXDEPTH];	/* depth changes on setcontext */

/*
 * The process ID of the first Shade process.  Trace output from all other
 * processes will be labeled as coming from a child process.
 */
static pid_t	pidFirst;

/*
 * This flag is set to TRUE if the application has more threads than we
 * can track.
 */
static spix_bool_t	isThdOvfl = FALSE;

const shade_options_t shadeuser_opts[]={
    {0, "", NULL, NO_PARAM}
}; 


static void 	flushwindows(unsigned);
static void	clear_stats(void);
static void 	pre_ticc(shade_trace_t *, const shade_regs_t *);
static void 	post_ticc(shade_trace_t *, const shade_regs_t *);
static void 	post_save(shade_trace_t *, const shade_regs_t *);
static void 	saverestrun(unsigned, spix_sparc_iop_t);



/*
**  Perform initializations specific to the 'window' analyzer.  Return
**  zero on success, non-zero on failure.
*/
int shadeuser_initialize(
	int     	argc,
	char ** 	argv,
	char ** 	envp)
{
	shade_iset_t *	piset;
	shade_tset_t *	ptset;
	spix_bool_t	ok;
	int      	i, j;
    int                 window_optnum;
    shade_option_val_t  window_opt_value;


	/*
	 * No options are supported.
	 */
	ok = TRUE;
	for (i = j = 1;  i < argc;  i++) {
        
        window_optnum = shade_getoptv(argc, argv, &i, &j, shadeuser_opts, &window_opt_value);
    
        if (window_optnum!=-1){
            shade_error("Unknown option '%s'.", argv[i]);
            ok = FALSE;
        }
	}

	if (!ok) {
		shade_usage();
		return(1);
	}

	/*
	 * Record the initial process ID.
	 */
	pidFirst = getpid();

	/*
	 * We want to be notified of fork() operations, so we know if the
	 * application spawns a child process.
	 */
	shade_setopt(SHADE_OPT_FORKNOTIFY);

	shade_trsize(sizeof(shade_trace_t));

	/*
	 * Trace the opcode, thread ID, and call 'post_save' for all SAVE
	 * instructions.
	 */
	piset = shade_iset_newop(SPIX_SPARC_IOP_SAVE, -1);
	ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID, -1);
	shade_trctl(shade_iset_newcopy(piset), SHADE_TRI_EXECUTED, ptset);
	shade_trfun(piset, SHADE_TRI_EXECUTED, 0, post_save, 0, 0);

	/*
	 * Trace the opcode and thread ID for all RESTORE instructions.
	 */
	piset = shade_iset_newop(SPIX_SPARC_IOP_RESTORE, SPIX_SPARC_IOP_RETURN,
				 -1);
	ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID, -1);
	shade_trctl(piset, SHADE_TRI_EXECUTED, ptset);

	/*
	 * Trace the opcode, thread ID, trap value, and taken flag for all
	 * trap instructions.  Also, call 'pre_ticc' and 'post_ticc'.
	 */
	piset = shade_iset_newclass(SHADE_ICLASS_TRAP, -1);
	ptset = shade_tset_new(SHADE_TRCTL_IH, SHADE_TRCTL_TID,
			       SHADE_TRCTL_TAKEN, SHADE_TRCTL_EA, -1);
	shade_trctl(shade_iset_newcopy(piset), SHADE_TRI_EXECUTED, ptset);
	shade_trfun(piset, SHADE_TRI_EXECUTED, pre_ticc, post_ticc, 0, 0);

	return(0);
}


/*
**  Run and trace an application that is loaded into Shade.
*/
int shadeuser_analyze(void)
{
	Wcntr *         	pwc;
	shade_trace_t   	atr[8192];
	shade_status_t  	status;
	spix_addr_t     	fps[MAXDEPTH][MAXTHDS];
	spix_addr_t     	nfp;
	unsigned        	depth[MAXTHDS];
	unsigned        	d;
	unsigned        	nwins;
	unsigned        	ntr;
	unsigned        	itr;
	unsigned        	tid;
	unsigned        	i;

	static pid_t           	pid_parent;

	pid_t           	pid;
	int             	val;


	status = shade_appstatus(&val);

	switch (status) {
	case SHADE_STATUS_LOADED:
		/*
		 * Initialize the emulated register window statistics for various
		 * register window sizes.
		 */
		for (i = 0;  i < NWCS;  i++) {
			wcs[i].wc_size = i + 1;
			for (tid = 0;  tid < MAXTHDS;  tid++)
				wcs[i].wc_depth[tid] = 1;
		}

		/*
		 * Initially, the application's window depth is zero.
		 */
		for (tid = 0;  tid < MAXTHDS;  tid++)
			depth[tid] = 0;

		/*
		 * Remember the parent process ID so we can tell if it forks.
		 */
		pid_parent = getpid();

	case SHADE_STATUS_FORKED:
		/*
		 * If the application forks, zero the statistics in the child.
		 * The child shouldn't inherit statistics from the parent.
		 */
		pid = getpid();
		if (pid != pid_parent) {
			pid_parent = pid;
			clear_stats();
		}
		break;
	}

	/*
	 * Emulate and trace the application.
	 */
	ntr = shade_run(atr, sizeof(atr)/sizeof(atr[0]));
	for (itr = 0;  itr < ntr;  itr++) {
		/*
		 * Get the ID of the thread producing this trace record.
		 */
		tid = atr[itr].tr_tid;
		if (tid >= MAXTHDS) {
			isThdOvfl = TRUE;
			continue;
		}

		switch (atr[itr].tr_ih)
		{
		case SPIX_SPARC_IOP_SAVE:
				/*
				 * Application allocates a new register
				 * window.  Check for consecutive SAVEs.
				 */
			saverestrun(tid, SPIX_SPARC_IOP_SAVE);
			d = ++(depth[tid]);
			nsaves++;

				/*
				 * Remember the %fp value so we know how
				 * many windows to restore on a setcontext.
				 */
			if (d < MAXDEPTH)
				fps[d][tid] = atr[itr].tr_ea;

				/*
				 * Calculate overflows for the various
				 * register window sizes.
				 */
			for (i = 0;  i < NWCS;  i++) {
				pwc = &wcs[i];
				if (pwc->wc_depth[tid] < pwc->wc_size)
					pwc->wc_depth[tid]++;
				else
					pwc->wc_nover++;
			}
			break;

		case SPIX_SPARC_IOP_RESTORE:
		case SPIX_SPARC_IOP_RETURN:
				/*
				 * Application restores to a previous
				 * register window.  Check for consecutive
				 * RESTOREs.
				 */
			saverestrun(tid, SPIX_SPARC_IOP_RESTORE);
			d = --(depth[tid]);
			nrestores++;

				/*
				 * Calculate underflows for the various
				 * register window sizes.
				 */
			for (i = 0;  i < NWCS;  i++) {
				if (wcs[i].wc_depth[tid] > 1)
					wcs[i].wc_depth[tid]--;
				else
					wcs[i].wc_nunder++;
			}
			break;

		case TICC_FLUSHWINDOWS:
				/*
				 * Application forces all register windows
				 * to be flushed.
				 */
			nflushwindows++;
			flushwindows(tid);
			continue;

		case TICC_SETCONTEXT:
				/*
				 * Application restores to a previous
				 * context.  This forces register windows to
				 * be flushed.
				 */
			nsetcontexts++;
			flushwindows(tid);

				/*
				 * Break any consecutive sequences of
				 * SAVEs or RESTOREs.
				 */
			saverestrun(tid, SPIX_SPARC_NIOP);

				/*
				 * Determine how many register windows are
				 * restored.
				 */
			nwins = 0;
			nfp = atr[itr].tr_ea;
			d = depth[tid];
			for (; d > 0 && nfp > fps[d][tid];  d--)
				nwins++;
			depth[tid] = d;
			if (nwins < MAXDEPTH)
				setcontexts[nwins]++;
			continue;

		default:
			continue;
		}

		/*
		 * Count the number of times a SAVE or RESTORE
		 * transitions to each window depth.
		 */
		if (d < MAXDEPTH)
			depthcnts[d]++;

		/*
		 * Keep track of the maximum window depth.
		 */
		if (maxdepth < d)
			maxdepth = d;
		sumdepth += d;
	}
	return(0);
}


void
shadeuser_report(int reason, void* ptr) {
	pid_t     	pid;
	unsigned        tid; 
	unsigned	cum;
	unsigned	cum2;
	unsigned	tot;
	unsigned	i;

	switch (reason) {
	case SHADEUSER_REPORT_TERMINATE:
	case SHADEUSER_REPORT_SIGNAL:		
		fprintf(shade_fp, "\n\n");
		/*
		 * If this is a child process, say so in the output.
		 */
		pid = getpid();
		if (pid != pidFirst)
			fprintf(shade_fp, "FROM CHILD PROCESS %d:\n\n", pid);

		/*
		 * Warn the user if data was lost due to thread overflow.
		 */
		if (isThdOvfl) {
			fprintf(shade_fp, "Warning: Application has too many "
				"threads, can only analyze first %d threads.\n\n",
				MAXTHDS);
		}

		/*
		 * Write the number of flush window traps and the number of
		 * setcontext system calls.  Both of these cause the application's
		 * register windows to be flushed.
		 */
		fprintf(shade_fp, "%u flushwindow traps\n", nflushwindows);
		fprintf(shade_fp, "%u setcontext syscalls\n", nsetcontexts);

		/*
		 * If there are no SAVE or RESTORE instructions, the remaining
		 * statistics are meaningless.
		 */
		if (!nsaves || !nrestores) {
			fflush(shade_fp);
			return;
		}

		tot = nsaves + nrestores;

		/*
		 * Write the number of overflows and underflows for each emulated
		 * register window size.
		 */
		fprintf(shade_fp, "\n\nREGISTER WINDOW OVERFLOWS/UNDERFLOWS:\n\n");

		fprintf(shade_fp, "%-8s%13s%11s%13s%11s%13s\n\n", "#windows", 
			"#overflows", "%overflow",
			"#underflows", "%underflow", "#flushed");

		for (i = 0;  i < NWCS;  i++) {
			if (wcs[i].wc_nover || wcs[i].wc_nunder)
				fprintf(shade_fp,
					"1+%-6d%13u%10.4f%%%13u%10.4f%%%13u\n",
					wcs[i].wc_size, 
					wcs[i].wc_nover,
					percent(wcs[i].wc_nover, nsaves),
					wcs[i].wc_nunder,
					percent(wcs[i].wc_nunder, nrestores),
					wcs[i].wc_nflush);
		}

		/*
		 * Write the application's maximum and average save depth.
		 */
		fprintf(shade_fp, "\n\nWINDOW SAVE DEPTH:\n\n");

		fprintf(shade_fp, "%-10u maximum save depth\n", maxdepth);
		fprintf(shade_fp, "%-10.4f average save depth\n",
			(double) sumdepth / tot);

		/*
		 * Write the number of times the application reached each window
		 * depth by executing a SAVE or RESTORE.
		 */
		fprintf (shade_fp, "\n%-5s%9s%11s%11s\n\n", "depth", "#times",
			 "%times", "cum%");

		cum = 0;
		for (i = 0;  i < MAXDEPTH;  i++) {
			if (depthcnts[i]) {
				cum += depthcnts[i];
				fprintf(shade_fp, "%-5u%9u%10.4f%%%10.4f%%\n", i,
					depthcnts[i], percent (depthcnts[i], tot),
					percent (cum, tot));
			}
		}


		/*
		 * Write statistics for the number of consecutive SAVE or RESTORE
		 * operations.
		 */
		fprintf(shade_fp, "\n\nWINDOW SAVE/RESTORE RUNS:\n\n");
		fprintf(shade_fp, "%-10u maximum save run length\n", maxsaverun);
		fprintf(shade_fp, "%-10.4f average save run length\n",
			(double) sumsaverun / nsaveruns);
		fprintf(shade_fp, "%-10u maximum restore run length\n", maxrestrun);
		fprintf(shade_fp, "%-10.4f average restore run length\n",
			(double) sumrestrun / nrestruns);

		fprintf (shade_fp, "\n%-6s%10s%11s%11s%10s%11s%11s", 
			 "run", "#save", "%save", "", "#restore", "%restore", "");
		fprintf (shade_fp, "\n%-6s%10s%11s%11s%10s%11s%11s\n\n", 
			 "length", "runs", "runs", "cum%", "runs", "runs", "cum%");

		cum = cum2 = 0;
		for (i = 0;  i < MAXRUN;  i++) {
			if (saveruncnts[i] || restruncnts[i]) {
				cum += saveruncnts[i];
				cum2 += restruncnts[i];
				fprintf(shade_fp, 
					"%-6d%10u%10.4f%%%10.4f%%%10u%10.4f%%%10.4f%%\n",
					i,
					saveruncnts[i], percent (saveruncnts[i], nsaveruns),
					percent (cum, nsaveruns),
					restruncnts[i], percent (restruncnts[i], nrestruns),
					percent (cum2, nrestruns));
			}
		}

		/*
		 * Write the statistics for the setcontext() system calls that
		 * executed.
		 */
		if (nsetcontexts > 0) {
			fprintf(shade_fp, "\n\nSETCONTEXT SYSTEM CALLS:\n\n");
			fprintf(shade_fp, "%-7s%8s%11s%11s\n\n",
				"#frames", "#calls", "%calls", "cum%");
			cum = 0;
			for (i = 0;  i < MAXDEPTH;  i++) {
				if (setcontexts[i]) {
					cum += setcontexts[i];
					fprintf(shade_fp, "%-7d%8u%10.4f%%%10.4f%%\n",
						i, setcontexts[i], 
						percent(setcontexts[i], nsetcontexts), 
						percent(cum, nsetcontexts));
				}
			}
		}

		fflush(shade_fp);

		break;
	  
	case SHADEUSER_REPORT_APP_DONE:	
		/*
		 * The application has ended.  Terminate the final sequence of
		 * consecutive SAVEs or RESTOREs.
		 */
		for (tid = 0;  tid < MAXTHDS;  tid++)
			saverestrun(tid, SPIX_SPARC_NIOP);		
		break;
	}
}

/*
**  Adjust the register window statistics in 'wcs' to reflect a register
**  window flush operation for the given thread.
*/
static void flushwindows(
	unsigned	tid)
{
	unsigned	i;


	assert(tid < MAXTHDS);

	for (i = 0;  i < NWCS;  i++) {
		wcs[i].wc_nflush += wcs[i].wc_depth[tid];
		wcs[i].wc_depth[tid] = 1;
	}
}


/*
**  Clear out all the statistics gathered so far.
*/
static void clear_stats(void)
{
	unsigned	i;


	/*
	 * Clear the counts of various events.
	 */
	nflushwindows = 0;
	nrestores = 0;
	nsaves = 0;
	nsaveruns = 0;

	/*
	 * Clear the register window statistics.
	 */
	for (i = 0;  i < NWCS;  i++) {
		wcs[i].wc_nover = 0;
		wcs[i].wc_nunder = 0;
		wcs[i].wc_nflush = 0;
	}

	/*
	 * Clear the window depth information and setcontext() information.
	 */
	maxdepth = 0;
	sumdepth = 0.0;
	nsetcontexts = 0;

	for (i = 0;  i < MAXDEPTH;  i++) {
		depthcnts[i] = 0;
		setcontexts[i] = 0;
	}

	/*
	 * Clear the statistics for consecutive SAVE and RESTORE instructions.
	 */
	maxrestrun = 0;
	maxsaverun = 0;
	sumrestrun = 0.0;
	sumsaverun = 0.0;

	for (i = 0;  i < MAXRUN;  i++) {
		restruncnts[i] = 0;
		saveruncnts[i] = 0;
	}

	/*
	 * It's no longer important if we lost data due to thread overflow.
	 */
	isThdOvfl = FALSE;
}


/*
**  User-trace function called before each trap (ticc) instruction.
*/
static void pre_ticc(
	shade_trace_t * 	ptr,
	const shade_regs_t *	psh)
{
	/*
	 * The 'tr_ea' field contains the trap value.  See if this is a
	 * register window flush trap.
	 */
	if (ptr->tr_ea == ST_FLUSH_WINDOWS) {
		ptr->tr_ih = TICC_FLUSHWINDOWS;
	}

	/*
	 * See if it is a setcontext() system call.
	 */
	else if ((ptr->tr_ea == ST_SYSCALL || ptr->tr_ea == 0x40 ) &&
		 psh->sh_r[REG_G1].ii[1] == SYS_context &&
		 psh->sh_r[REG_O0].ii[1] == 1)
	{
		ptr->tr_ih = TICC_SETCONTEXT;
	}
}


/*
**  User-trace function called after each trap (ticc) instruction.
*/
static void post_ticc(
	shade_trace_t * 	ptr,
	const shade_regs_t *	psh)
{
	if (ptr->tr_ih == TICC_FLUSHWINDOWS || ptr->tr_ih == TICC_SETCONTEXT) {
		if (!ptr->tr_taken) {
			/* 
			 * If the trap wasn't taken, change the 'tr_ih'
			 * field, so the code in shadeuser_analyze() doesn't
			 * emulate the effect of the trap on the register
			 * windows.
			 */
			ptr->tr_ih = SPIX_SPARC_NIOP;
		}
		else {
			/*
			 * Record the new frame pointer value set by
			 * setcontext() system calls.  We need this to
			 * figure out how many register window levels are
			 * restored.
			 */
			ptr->tr_ea = psh->sh_r[REG_FP].ux;
		}
	}
}


/*
**  User-trace function called after each SAVE instruction.
*/
static void post_save(
	shade_trace_t * 	ptr,
	const shade_regs_t *	psh)
{
	/*
	 * Record the frame pointer for this new register window depth.
	 */
	ptr->tr_ea = psh->sh_r[REG_FP].ux;
}


/*
**  Perform termination activities after all applications have been executed.
*/


int shadeuser_terminate(
	int	ret)
{
	int	val;

	/*
	 * Exit with the termination status from the executed application.
	 */
	switch (shade_appstatus(&val))
	{
	case SHADE_STATUS_EXITED:
		return(val);

	case SHADE_STATUS_SIGNALED:
		return(128 + val);

	case SHADE_STATUS_NOAPP:
		return(0);

	case SHADE_STATUS_LOADED:
	case SHADE_STATUS_RUNNING:
	default:
		assert(0);
	}

}


/*
**  Print a usage message for the command line options that are specific
**  to the 'window' analyzer.
*/
void shadeuser_analusage(void)
{
    shade_print_opt_info(shadeuser_opts, 0);
}


/*
**  Count the number of consecutive SAVE or RESTORE instructions for the
**  given thread.
*/
static void saverestrun(
	unsigned        	tid,
	spix_sparc_iop_t	ih)
{
	static unsigned 	run[MAXTHDS];
	static spix_sparc_iop_t	lastih[MAXTHDS];

	unsigned	r;


	assert(tid < MAXTHDS);

	if (!run[tid]) {
		lastih[tid] = ih;
		run[tid] = 1;
		return;
	}

	if (ih == lastih[tid]) {
		run[tid]++;
		return;
	}

	r = run[tid];

	switch (lastih[tid])
	{
	case SPIX_SPARC_IOP_SAVE:
		nsaveruns++;
		sumsaverun += r;
		if (maxsaverun < r)
			maxsaverun = r;
		if (r < MAXRUN)
			saveruncnts[r]++;
		break;

	case SPIX_SPARC_IOP_RESTORE:
		nrestruns++;
		sumrestrun += r;
		if (maxrestrun < r)
			maxrestrun = r;
		if (r < MAXRUN)
			restruncnts[r]++;
		break;
	}

	lastih[tid] = ih;
	run[tid] = 1;
}
