/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  mpshade.h	- Interface to the multi-process Shade library
**
**	MPShade is multi-process extension that layers on top of Shade.
**	A single MPShade analyzer process can collect trace data from many
**	MPShade tracer processes.  Each tracer runs a single application
**	under Shade and feeds trace records to the shared analyzer.
**
**	This header defines the processor independent interfaces to
**	MPShade.  Analyzers and tracers that are not dependent on the
**	target processor can include just this header.  However, they must
**	define the SPIX_ADDR macro to define the size (in bits) of a
**	target address.  Analyzers and tracers that require processor
**	specific information must include the appropriate processor
**	specific MPShade header.  Since the processor specific header
**	defines the target address size, they need not define SPIX_ADDR.
*/

#ifndef MPSHADE_H
#define MPSHADE_H

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stddef.h>
#include <sys/types.h>
#include <shade.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * The type of a tracer ID.
 */
typedef unsigned	mpshade_id_t;

#define MPSHADE_ANYID  	((mpshade_id_t)-2)


/*
 * A return value indicating an error.
 */
#define MPSHADE_ERROR	(~0U)


/*
 * The possible event types returned to an analyzer using the primitive
 * interfaces.
 */
typedef enum {
	MPSHADE_CREATE      = (1<<0),
	MPSHADE_FORK        = (1<<1),
	MPSHADE_TRACEBUF    = (1<<2),
	MPSHADE_USER        = (1<<3),
	MPSHADE_TERMINATE   = (1<<4),
	MPSHADE_FINISHED    = (1<<5),
	MPSHADE_EXECTRACE   = (1<<6),
	MPSHADE_EXECNOTRACE = (1<<7),

	/*
	 * A mask selecting no events.
	 */
	MPSHADE_NOEVENT = 0,

	/*
	 * A mask selecting all events.
	 */
	MPSHADE_ANYEVT  = (MPSHADE_CREATE    | MPSHADE_FORK | 
	                   MPSHADE_TRACEBUF  | MPSHADE_USER | 
	                   MPSHADE_TERMINATE | MPSHADE_FINISHED |
			   MPSHADE_EXECTRACE | MPSHADE_EXECNOTRACE)
} mpshade_etype_t;


/*
 * The definition of an event returned to an analyzer using the primitive
 * interfaces.
 */
typedef struct {
	/*
	 * The ID of the tracer reporting the event.
	 */
	mpshade_id_t	id;

	union {
		/*
		 * Data for an MPSHADE_CREATE, MPSHADE_EXECTRACE, or
		 * MPSHADE_EXECNOTRACE event.
		 */
		struct {
			char *  	path;	/* application pathname */
			pid_t   	pid;	/* process ID of tracer */
			int      	argc;	/* application arg count */
			char ** 	argv;	/* application arguments */
			char ** 	envp;	/* application environment */
			char *  	path_interp;	/* interpreter path */
			spix_addr64_t	addr_interp;	/* base address */
		} create;

		/*
		 * Data for an MPSHADE_FORK event.
		 */
		struct {
			mpshade_id_t	trid_parent;	/* ID of parent */
			pid_t   	pid;    	/* pid of tracer */
		} fork;

		/*
		 * Data for an MPSHADE_TRACEBUF event.
		 */
		struct {
			size_t   	ntrace; 	/* no. trace records */
			size_t   	sztrace;	/* size of record */
			shade_trace_t *	ptrace; 	/* array of records */
		} tracebuf;

		/*
		 * Data for an MPSHADE_USER event.
		 */
		struct {
			size_t      	len;    	/* size of data */
			void *  	pbuf;   	/* the data */
		} user;

		/*
		 * Data for an MPSHADE_TERMINATE event.
		 */
		struct {
			shade_status_t	status; 	/* status of app */
			int         	code;   	/* exit val or signum */
		} terminate;

		char	reserve[64];
	} data;
} mpshade_evt_t;


/*
 * Information returned from mpshade_appinfo().
 */
typedef struct {
	pid_t   	pid;	/* process ID */
	int     	code;	/* exit code or terminating signal */
	int     	argc;	/* argument count */
	const char *	path;	/* application pathname */
	char * const *	argv;	/* argument vector */
	char * const *	envp;	/* environment vector */
} mpshade_appinfo_t;


/*
 * Possible status values returned from mpshade_appinfo().
 */
typedef enum {
	MPSHADE_STATUS_NOAPP,
	MPSHADE_STATUS_RUNNING,
	MPSHADE_STATUS_EXITED,
	MPSHADE_STATUS_SIGNALED,
	MPSHADE_STATUS_EXECED
} mpshade_status_t;


/*
 * Possible flag values for mpshade_evt_waitid() and mpshade_setrunflags().
 */
typedef enum {
	MPSHADE_NOBLOCK  = (1<<0),
	MPSHADE_NOFINISH = (1<<1)
} mpshade_waitflag_t;


/*
 * Possible trace modes for mpshade_setmode().
 */
typedef enum {
	MPSHADE_SUSPENDED,
	MPSHADE_TRACING
} mpshade_mode_t;


/*
 * Possible trace record sizes for mpshade_trsize().
 */
typedef enum {
	MPSHADE_TRSZ_TRACE	= 0,
	MPSHADE_TRSZ_TRACEID	= 1
} mpshade_trsz_t;


/*
 * The definition of a trace record for MPSHADE_TRSZ_TRACE.
 */
#define MPSHADE_TRACE   	\
	SHADE_TRACE


/*
 * The definition of a trace record for MPSHADE_TRSZ_TRACEID.
 */
#define MPSHADE_TRACEID 	\
	SHADE_TRACE     	\
	mpshade_id_t	tr_trid;


typedef struct mpshade_iset_s	mpshade_iset_t;
typedef struct mpshade_tset_s	mpshade_tset_t;


void		mpshadeuser_usage();

/*
**  Primitive layer analyzer interfaces.
*/
int             	mpshadeuser_analyzer(int, char **, char **);
mpshade_id_t     	mpshade_execv(const char *, const char *,
			    char * const *);
mpshade_id_t     	mpshade_execl(const char *, ...);
int              	mpshade_send_msg(mpshade_id_t, size_t, const void *);
mpshade_etype_t 	mpshade_evt_wait(mpshade_evt_t *);
mpshade_etype_t 	mpshade_evt_waitid(mpshade_id_t, mpshade_etype_t,
                 	    mpshade_waitflag_t, mpshade_evt_t *);
int             	mpshade_splitargs(char **, char ***, int *);

/*
**  Primitive layer tracer interfaces.
*/
int     	mpshadeuser_tracer(int, char **, char **);
int     	mpshade_report_trace(size_t, const shade_trace_t *);
int     	mpshade_report_user(size_t, const void *);
int     	mpshade_get_msg(size_t *, void **);
int     	mpshade_attach(pid_t);
mpshade_id_t 	mpshade_myid(void);
int     	mpshade_finishanal(void);


/*
**  Combined-trace layer analyzer interfaces.
*/
int              	mpshadeuser_combined(int, char **, char **);
mpshade_id_t     	mpshade_ntracers(void);
mpshade_status_t 	mpshade_appinfo(mpshade_id_t, mpshade_appinfo_t *);
int              	mpshade_setcallback(mpshade_etype_t,
                 	    void (*)(mpshade_etype_t, const mpshade_evt_t *,
                 	        void *),
                 	    void *);
int              	mpshade_setrunflags(mpshade_waitflag_t);
void             	mpshade_setslice(size_t);
shade_trace_t *  	mpshade_stepid(mpshade_id_t *);
shade_trace_t *  	mpshade_step(void);
size_t           	mpshade_runid(shade_trace_t *, size_t, mpshade_id_t *);
size_t           	mpshade_run(shade_trace_t *, size_t);


/*
**  Standard-tracer layer analyzer interfaces.
*/
int		mpshadeuser_stdtr(int, char **, char **);
mpshade_id_t 	mpshade_load(const char *, char * const *, char * const *);
int     	mpshade_trsize(mpshade_trsz_t);
mpshade_iset_t *mpshade_iset_newclass(shade_iclass_t, ...);
mpshade_iset_t *mpshade_iset_newcopy(mpshade_iset_t *);
void    	mpshade_iset_free(mpshade_iset_t *);
mpshade_iset_t *mpshade_iset_addclass(mpshade_iset_t *, shade_iclass_t, ...);
mpshade_tset_t *mpshade_tset_new(shade_trctl_t, ...);
mpshade_tset_t *mpshade_tset_newcopy(mpshade_tset_t *);
void    	mpshade_tset_free(mpshade_tset_t *);
mpshade_tset_t *mpshade_tset_add(mpshade_tset_t *, shade_trctl_t, ...);
int     	mpshade_trctl(mpshade_iset_t *, shade_tri_t, mpshade_tset_t *);
int     	mpshade_setmode(mpshade_mode_t);

#define mpshade_analyzer mpshadeuser_analyzer
#define mpshade_stdtr    mpshadeuser_stdtr
#define mpshade_combined mpshadeuser_combined
#define mpshade_tracer   mpshadeuser_tracer

#ifdef __cplusplus
}
#endif

#endif /*MPSHADE_H*/

